# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.selective_backprop.selective_backprop import SelectiveBackprop as SelectiveBackprop
from composer.algorithms.selective_backprop.selective_backprop import \
    SelectiveBackpropHparams as SelectiveBackpropHparams



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import atexit
import sys
from typing import Any

from composer.core.logging import LogLevel, RankZeroLoggerBackend, TLogData
from composer.core.types import Logger, State, StateDict

import wandb  # isort:skip


class WandBLoggerBackend(RankZeroLoggerBackend):
    """Log to Weights and Biases (https://wandb.ai/)

    Args:
        kwargs (Any): Parameters to pass into :meth:`wandb.init`.
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__()
        self._init_params = kwargs

    def _log_metric(self, epoch: int, step: int, log_level: LogLevel, data: TLogData):
        wandb.log(data, step=step)

    def _training_start(self, state: State, logger: Logger) -> None:
        wandb.init(**self._init_params)
        atexit.register(self._close_wandb)

    def state_dict(self) -> StateDict:
        # Storing these fields in the state dict to support run resuming in the future.
        return {"name": wandb.run.name, "project": wandb.run.project, "entity": wandb.run.entity, "id": wandb.run.id}

    def _close_wandb(self) -> None:
        # wandb.finish isn't automatically called during ddp-spawn, so manually calling it on rank 0
        exc_tpe, exc_info, tb = sys.exc_info()
        if (exc_tpe, exc_info, tb) == (None, None, None):
            wandb.finish(0)
        else:
            # record there was an error
            wandb.finish(1)



# Copyright 2021 MosaicML. All Rights Reserved.

from dataclasses import asdict, dataclass

import numpy as np
import torch
import yahp as hp
from PIL import Image
from PIL.Image import Image as ImageType

from composer.algorithms.algorithm_hparams import AlgorithmHparams
from composer.core.event import Event
from composer.core.types import Algorithm, Event, List, Logger, State
from composer.utils.augmentation_primitives import augmentation_sets
from composer.utils.data import add_dataset_transform


@dataclass
class AugMixHparams(AlgorithmHparams):

    severity: int = hp.optional(doc="Intensity of each augmentation. Ranges from 0 (none) to 10 (maximum)", default=3)
    depth: int = hp.optional(doc="Number of augmentations to compose in a row", default=-1)
    width: int = hp.optional(doc="Number of parallel augmentation chains to combine", default=3)
    alpha: float = hp.optional(doc="Mixing parameter for clean vs. augmented images.", default=1.0)
    augmentation_set: str = hp.optional(
        doc=
        "Set of augmentations to sample from. 'all', 'safe' (only augmentations that don't appear on CIFAR10C/ImageNet10C), or 'original'",
        default="all")

    def initialize_object(self) -> "AugMix":
        return AugMix(**asdict(self))


def augment_and_mix(img: ImageType = None,
                    severity: int = 3,
                    depth: int = -1,
                    width: int = 3,
                    alpha: float = 1.0,
                    augmentation_set: List = augmentation_sets["all"]) -> ImageType:
    """
    Perform augmentations.
    """

    assert isinstance(img, ImageType) or isinstance(img, np.ndarray), "img must be a PIL.Image"
    chain_weights = np.float32(np.random.dirichlet([alpha] * width))
    mixing_weight = np.float32(np.random.beta(alpha, alpha))
    augmented_combination = np.zeros_like(img, dtype=np.float32)

    # Iterate over image chains
    for chain_i in range(width):
        augmented_image = img.copy()
        # Determine depth of current augmentation chain
        if depth > 0:
            d = depth
        else:
            d = np.random.randint(1, 4)
        # Iterate through chain depth
        for _ in range(d):
            aug = np.random.choice(augmentation_set)
            augmented_image = aug(augmented_image, severity)
        augmented_combination += chain_weights[chain_i] * np.asarray(augmented_image)
    mixed = (1 - mixing_weight) * np.asarray(img) + mixing_weight * augmented_combination
    mixed = Image.fromarray(np.uint8(mixed))
    return mixed


class AugmentAndMixTransform(torch.nn.Module):

    def __init__(self,
                 severity: int = 3,
                 depth: int = -1,
                 width: int = 3,
                 alpha: float = 1.0,
                 augmentation_set: str = "all"):
        """
        See documentation for RandAugment
        """
        super().__init__()
        if severity < 0 or severity > 10:
            raise ValueError("AugMix severity value must be 0 ≤ severity ≤ 10")
        if width < 1:
            raise ValueError("AugMix width must be ≥ 1")
        if augmentation_set not in augmentation_sets.keys():
            raise KeyError(f"AugMix augmentation_set is not one of {augmentation_sets.keys()}")
        self.severity = severity
        self.depth = depth
        self.width = width
        self.alpha = alpha
        self.augmentation_set = augmentation_sets[augmentation_set]

    def forward(self, img: ImageType) -> ImageType:

        return augment_and_mix(img=img,
                               severity=self.severity,
                               depth=self.depth,
                               width=self.width,
                               alpha=self.alpha,
                               augmentation_set=self.augmentation_set)


class AugMix(Algorithm):
    """
    Object that does AugMix (Hendrycks et al. (2020), AugMix: A Simple Data
        Processing Method to Improve Robustness and Uncertainty). Can be passed as a
        transform to torchvision.transforms.Compose().

    Args:
        severity: Severity of augmentation operators (between 1 to 10).
        width: Width of augmentation chains (number of parallel augmentations)
        depth: Depth of each augmentation chain. -1 enables stochastic depth
            uniformly from [1, 3]
        alpha: Probability coefficient for Beta and Dirichlet distributions. Sampling
            from Dirichlet determines relative weights of each augmented image
            sampling from beta determines relative weights of unaugmented and
            augmented images.
        augmentation_set: String, one of ["augmentations_all",
            "augmentations_corruption_safe", "augmentations_original"]. Set of
            augmentations to use. "augmentations_corruption_safe" excludes transforms
            that are part of the ImageNet-C/CIFAR10-C test sets.
            "augmentations_original" uses all augmentations, but some of the
            implementations are identical to the original github repo, which appears
            to contain implementation specificities for the augmentations "color",
            "contrast", "sharpness", and "brightness".
    """

    def __init__(self,
                 severity: int = 3,
                 depth: int = -1,
                 width: int = 3,
                 alpha: float = 1.0,
                 augmentation_set: str = "all"):
        if severity < 0 or severity > 10:
            raise ValueError("AugMix severity value must be 0 ≤ severity ≤ 10")
        if width < 1:
            raise ValueError("AugMix width must be ≥ 1")
        if augmentation_set not in augmentation_sets.keys():
            raise KeyError(f"AugMix augmentation_set is not one of {augmentation_sets.keys()}")
        self.hparams = AugMixHparams(severity=severity,
                                     depth=depth,
                                     width=width,
                                     alpha=alpha,
                                     augmentation_set=augmentation_set)

    def match(self, event: Event, state: State) -> bool:
        """ Runs on Event.TRAINING_START
        """
        return event == Event.TRAINING_START

    def apply(self, event: Event, state: State, logger: Logger) -> None:
        """ Inserts AugMix into the list of dataloader transforms
        """
        am = AugmentAndMixTransform(**self.hparams.to_dict())
        assert state.train_dataloader is not None, "Train Dataloader is not initialized."
        dataset = state.train_dataloader.dataset
        add_dataset_transform(dataset, am)



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import logging
import time
import warnings
from dataclasses import dataclass
from typing import List, Sequence

import yahp as hp

from composer import Logger, State
from composer.callbacks.callback_hparams import CallbackHparams
from composer.core.callback import Callback
from composer.core.types import BreakEpochException

log = logging.getLogger(__name__)


class Benchmarker(Callback):
    """Fast-forward the training loop to record 
    throughput for specific epochs and/or steps.

    It modifies the :attr:`~composer.core.State.step` and
    :attr:`~composer.core.State.epoch` to fast-forward the training loop,
    so that algorithms that activate at specific times will trigger
    and can be profiled.

    It stops the training process after the last step or epoch is profiled.

    It logs:

    * The throughput (averaged over the number of steps profiled in an epoch)
      to the ``throughput/step`` key.
    * The total elapsed training time to the ``wall_clock_train`` key.

    .. warning::
        The :class:`Benchmarker`: should NOT be used in conjunction with the
        :class:`~composer.callbacks.speed_monitor.SpeedMonitor`, since they
        log to the same keys.

    .. warning::
        The :class:`Benchmarker`: modifies the :class:`~compose.core.State`,
        which is an exception to the convention that callbacks should NOT
        modify state. This callback may break other algorithms and callbacks.

    Args:
        min_steps (int, optional):
            Maximum number of steps to profile per epoch, regardless of the length of
            regardless of the length of :attr:`step_list`.
            Defaults to 50.
        epoch_list (Sequence[int], optional).
            List of epochs at which to measure throughput.
            Defaults to [0, 1].
        step_list (Sequence[int], optional).
            List of steps at which to measure throughput.
            Defaults to [0, 50].
        all_epochs (bool, optional).
            Whether to override epoch_list and profile at all epochs.
    
            If False (the default), then it fast-forwards to
            the steps and epochs being profiled (specified by :attr:`epoch_list`
            and :attr:`step_list`, respectively).

            Otherwise, if True, then the throughput for
            the first :attr:`min_steps` batches of every epoch are recorded.
    """

    def __init__(self,
                 min_steps: int = 50,
                 epoch_list: Sequence[int] = (0, 1),
                 step_list: Sequence[int] = (0, 50),
                 all_epochs: bool = False):
        super().__init__()
        self.hparams = BenchmarkerHparams(min_steps=min_steps,
                                          epoch_list=list(epoch_list),
                                          step_list=list(step_list),
                                          all_epochs=all_epochs)
        if not all_epochs:
            if len(epoch_list) == 0:
                raise ValueError("'epoch_list'  must be non-empty.")
            if 0 not in epoch_list:
                raise ValueError("'epoch_list' must contain 0, otherwise first K epochs have unknown speed")
        if len(step_list) == 0:
            raise ValueError("'step_list'  must be non-empty.")
        if 0 not in step_list:
            raise ValueError("'step_list' must contain 0 because `EPOCH_START` requires batch_idx of 0")

        # Sort lists so that time only moves forward
        epoch_list = list(sorted(epoch_list))
        step_list = list(sorted(step_list))

        self.current_time = None
        self.profile_examples = 0
        self.profile_steps = 0
        self.profile_time = 0
        self.wall_clock_train = 0

        self.min_steps = min_steps

        self.all_epochs = all_epochs
        self.epoch_list = epoch_list
        self.epoch_ix = 0
        self.step_list = step_list
        self.step_ix = 0

        # initialized at training_start
        self.original_max_epochs = -1
        self.wct_dict = {}

    def _compute_elapsed_wct(self, epoch_wct_dict, steps_per_epoch: int, n_epochs: int):
        wct = 0.0
        wct_per_step = 0
        assert 0 in epoch_wct_dict, "epoch_wct_dict must contain 0"
        for step in range(steps_per_epoch):
            if step in epoch_wct_dict:
                wct_per_step = epoch_wct_dict[step]
            wct += wct_per_step
        return wct * n_epochs

    def training_start(self, state: State, logger: Logger):
        del logger  # Unused
        warnings.warn("The timing monitor is activated. The model will not be fully trained."
                      "All quality metrics for this run will be incorrect.")
        self.wall_clock_train = 0.0
        self.original_max_epochs = state.max_epochs
        # maybe override epoch_list
        if self.all_epochs:
            self.epoch_list = list(range(state.max_epochs))
            log.info(f"all_epochs=True, overriding epoch_list to be every epoch from 0 to {state.max_epochs}")
        self.wct_dict = {e: {s: -1.0 for s in self.step_list} for e in self.epoch_list}
        state.max_epochs = len(self.epoch_list)

    def epoch_end(self, state: State, logger: Logger):
        prev_epoch = self.epoch_list[self.epoch_ix]
        epoch_wct_dict = self.wct_dict[prev_epoch]
        self.epoch_ix += 1
        if self.epoch_ix < len(self.epoch_list):
            next_epoch = self.epoch_list[self.epoch_ix]
        else:
            next_epoch = self.original_max_epochs

        state.epoch = next_epoch - 1
        state.step = next_epoch * state.steps_per_epoch
        n_epochs = next_epoch - prev_epoch

        self.wall_clock_train += self._compute_elapsed_wct(epoch_wct_dict, state.steps_per_epoch, n_epochs)
        logger.metric_epoch({'wall_clock_train': self.wall_clock_train})

    def batch_start(self, state: State, logger: Logger):
        del state, logger  # Unused
        if self.current_time is None:
            self.current_time = time.time()
            self.profile_examples = 0
            self.profile_steps = 0
            self.profile_time = 0.0

    def batch_end(self, state: State, logger: Logger):
        if self.current_time is not None:
            now = time.time()
            elapsed = now - self.current_time
            self.current_time = now
            self.profile_examples += state.last_batch_size * state.world_size
            self.profile_steps += 1
            self.profile_time += elapsed

            if self.profile_steps >= self.min_steps:
                avg_throughput = self.profile_examples / self.profile_time
                avg_time_per_step = self.profile_time / self.profile_steps
                profile_epoch = self.epoch_list[self.epoch_ix]
                profile_step = self.step_list[self.step_ix]
                self.wct_dict[profile_epoch][profile_step] = avg_time_per_step
                logger.metric_batch({'throughput/step': avg_throughput})

                self.current_time = None
                self.step_ix += 1
                if self.step_ix == len(self.step_list):
                    self.step_ix = 0
                    raise BreakEpochException
                else:
                    state.step = state.epoch * state.steps_per_epoch + self.step_list[self.step_ix]


@dataclass
class BenchmarkerHparams(CallbackHparams):
    """Parameters for the :class:`Benchmarker`

    See the documentation for the :class:`Benchmarker`.
    """
    min_steps: int = hp.optional(
        doc="Minimum number of steps to use for measuring throughput.",
        default=50,
    )
    epoch_list: List[int] = hp.optional(
        doc="List of epochs at which to measure throughput.",
        default_factory=lambda: [0, 1],
    )
    step_list: List[int] = hp.optional(
        doc="List of steps at which to measure throughput.",
        default_factory=lambda: [0, 50],
    )
    all_epochs: bool = hp.optional(
        doc="If true, override epoch_list and profile at all epochs.",
        default=False,
    )

    def initialize_object(self) -> Benchmarker:
        return Benchmarker(
            min_steps=self.min_steps,
            epoch_list=self.epoch_list,
            step_list=self.step_list,
            all_epochs=self.all_epochs,
        )



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.models.classify_mnist.mnist_hparams import MnistClassifierHparams as MnistClassifierHparams
from composer.models.classify_mnist.model import MNIST_Classifier as MNIST_Classifier

_task = 'Image Classification'
_dataset = 'MNIST'
_name = 'SimpleConvNet'
_quality = ''
_metric = 'Accuracy'
_ttt = '?'
_hparams = 'hparams_cpu.yaml'



# Copyright 2021 MosaicML. All Rights Reserved.

from dataclasses import dataclass
from typing import Callable, List, Optional, Sequence, Union

import torch
import torch.utils.data
import yahp as hp

from composer.datasets.hparams import DataloaderSpec, DatasetHparams
from composer.utils.string_enum import StringEnum


class SyntheticDataType(StringEnum):
    INCREASING = "increasing"
    GAUSSIAN = "gaussian"
    SEPARABLE = "separable"


class MemoryFormat(StringEnum):
    CONTIGUOUS_FORMAT = "contiguous_format"
    CHANNELS_LAST = "channels_last"
    CHANNELS_LAST_3D = "channels_last_3d"
    PRESERVE_FORMAT = "preserve_format"


class SyntheticDataset(torch.utils.data.Dataset):

    def __init__(self,
                 *,
                 sample_pool_size: int,
                 shape: Sequence[int],
                 memory_format: Union[str, MemoryFormat],
                 device: str,
                 one_hot: bool,
                 num_classes: int,
                 transform: Optional[Callable] = None,
                 data_type: SyntheticDataType = SyntheticDataType.GAUSSIAN):
        self.size = sample_pool_size
        self.shape = shape
        self.input_data = None
        self.input_target = None
        self.memory_format = getattr(torch, MemoryFormat(memory_format).value)
        self.device = device
        self.one_hot = one_hot
        self.num_classes = num_classes
        self.transform = transform
        self.data_type = data_type

    def __len__(self) -> int:
        return self.size

    def __getitem__(self, idx: int):
        if self.input_data is None:
            # Generating data on the first call to __getitem__ so that data is stored on the correct gpu,
            # after DeviceSingleGPU calls torch.cuda.set_device
            # This does mean that the first batch will be slower
            # generating samples so all values for the sample are the sample index
            # e.g. all(input_data[1] == 1). Helps with debugging.
            assert self.input_target is None
            if self.data_type == SyntheticDataType.GAUSSIAN or \
                self.data_type == SyntheticDataType.SEPARABLE:
                input_data = torch.randn(self.size, *self.shape)
            else:
                input_data = torch.arange(start=0, end=self.size, step=1, dtype=torch.float)
                input_data = input_data.reshape(self.size, *(1 for _ in self.shape))
                input_data = input_data.expand(self.size, *self.shape)  # returns a view
            assert input_data.shape == (self.size, *self.shape)
            input_data = torch.clone(input_data)  # allocate actual memory
            input_data = input_data.contiguous(memory_format=self.memory_format)
            input_data = input_data.to(self.device)
            if self.one_hot:
                input_target = torch.empty(self.size, self.num_classes).to(self.device)
                input_target[:, 0] = 1.0
            else:
                input_target = torch.randint(0, self.num_classes, (self.size,))

            # If separable, force the positive examples to have a higher mean than the negative examples
            if self.data_type == SyntheticDataType.SEPARABLE:
                assert not self.one_hot, "SyntheticDataType.SEPARABLE does not support one_hot=True."
                assert max(input_target) == 1 and min(input_target) == 0, \
                    "SyntheticDataType.SEPARABLE only supports binary labels"
                # Make positive examples have mean = 3 and negative examples have mean = -3
                # so they are easier to separate with a classifier
                input_data[input_target == 0] -= 3
                input_data[input_target == 1] += 3

            self.input_data = input_data
            self.input_target = input_target
        assert self.input_target is not None
        if self.transform is not None:
            return self.transform(self.input_data[idx]), self.input_target[idx]
        else:
            return self.input_data[idx], self.input_target[idx]


@dataclass
class SyntheticDatasetHparams(DatasetHparams):
    num_classes: int = hp.required("Number of classes")
    shape: List[int] = hp.required("Shape of tensor")
    one_hot: bool = hp.required("Whether to use one-hot encoding", template_default=False)
    device: str = hp.required(
        "Device to store the sample pool. "
        "Set to `cuda` to store samples on the GPU and eliminate PCI-e bandwidth with the dataloader. "
        "Set to `cpu` to move data between host memory and the gpu on every batch. ",
        template_default="cpu")
    memory_format: MemoryFormat = hp.optional("Memory format for the sample pool",
                                              default=MemoryFormat.CONTIGUOUS_FORMAT)
    sample_pool_size: int = hp.optional("Number of samples", default=100)
    drop_last: bool = hp.optional("Whether to drop the last samples for the last batch", default=True)
    shuffle: bool = hp.optional("Whether to shuffle the dataset for each epoch", default=True)
    data_type: SyntheticDataType = hp.optional("Type of synthetic data to create.", default=SyntheticDataType.GAUSSIAN)

    def initialize_object(self) -> DataloaderSpec:
        return DataloaderSpec(
            SyntheticDataset(
                num_classes=self.num_classes,
                shape=self.shape,
                one_hot=self.one_hot,
                device=self.device,
                memory_format=self.memory_format,
                sample_pool_size=self.sample_pool_size,
                data_type=self.data_type,
            ),
            drop_last=self.drop_last,
            shuffle=False,
        )



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import abc
from dataclasses import dataclass

import yahp as hp

from composer.core.callback import Callback


@dataclass
class CallbackHparams(hp.Hparams, abc.ABC):
    """Base class for :class:`~composer.core.callback.Callback` hyperparameters.
    
    Callback parameters that are added to
    :attr:`composer.trainer.trainer_hparams.TrainerHparams.callbacks`
    (e.g. via YAML or the CLI) are initialized in the training loop.
    """

    @abc.abstractmethod
    def initialize_object(self) -> Callback:
        """Initialize the callback.

        Returns:
            Callback: An instance of the callback.
        """
        pass



# Copyright 2021 MosaicML. All Rights Reserved.

from dataclasses import dataclass

from composer.models.model_hparams import ModelHparams


@dataclass
class UnetHparams(ModelHparams):

    def initialize_object(self):
        from composer.models.unet.unet import UNet
        return UNet(self)



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.models.base import MosaicClassifier
from composer.models.efficientnetb0.efficientnetb0_hparams import EfficientNetB0Hparams
from composer.models.efficientnets import EfficientNet


class EfficientNetB0(MosaicClassifier):

    def __init__(self, hparams: EfficientNetB0Hparams) -> None:
        model = EfficientNet.get_model_from_name(
            "efficientnet-b0",
            hparams.num_classes,
            hparams.drop_connect_rate,
        )
        super().__init__(module=model)



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import logging
from dataclasses import asdict, dataclass
from functools import partial
from typing import Optional, Tuple

import torch
import torch.nn.functional as F
import yahp as hp
from torchvision import transforms

from composer.algorithms import AlgorithmHparams
from composer.core import Algorithm, Event, Logger, State
from composer.core.types import Tensor

log = logging.getLogger(__name__)

_VALID_MODES = ("crop", "resize")


def resize_inputs(X: torch.Tensor,
                  y: torch.Tensor,
                  scale_factor: float,
                  mode: str = "resize",
                  resize_targets: bool = False) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Resize inputs and optionally outputs by cropping or interpolating.

    Args:
        X (torch.Tensor): Input tensor of shape (N, C, H, W). Resizing will be done along
                          dimensions H and W using the constant factor scale_factor.
        y (torch.Tensor): If resize_targets is True, output tensor of shape (N, C, H, W)
                          that will also be resized.
        scale_factor (float): Scaling coefficient for the height and width of the
                              input/output tensor. 1.0 keeps the original size.
        mode (str): Type of scaling to perform. Value must be one of 'crop' or 'resize'.
                    'crop' performs a random crop, whereas 'resize' performs a bilinear
                    interpolation. Default: 'crop'.
        resize_targets (bool): Resize the targets, y, as well. Default: False.

    Returns:
        X_sized (torch.Tensor): Resized input tensor of shape
                                (N, C, H * scale_factor, W * scale_factor).
        y_sized (torch.Tensor): If resized_targets is True, resized output tensor of shape
                                (N, C, H * scale_factor, W * scale_factor). Returns
                                original y, otherwise.
    """

    # Short-circuit if nothing should be done
    if scale_factor >= 1:
        return X, y

    def resize_tensor(tensor: torch.Tensor) -> torch.Tensor:
        # Reduce the size of input images, either via cropping or downsampling
        if mode.lower() == "crop":
            Hc = int(scale_factor * tensor.shape[2])
            Wc = int(scale_factor * tensor.shape[3])
            resize_transform = transforms.RandomCrop((Hc, Wc))
        elif mode.lower() == "resize":
            resize_transform = partial(F.interpolate, scale_factor=scale_factor, mode='bilinear')
        else:
            raise ValueError(f"Progressive mode '{mode}' not supported.")
        return resize_transform(tensor)

    X_sized = resize_tensor(X)
    if resize_targets:
        y_sized = resize_tensor(y)
    else:
        y_sized = y
    return X_sized, y_sized


@dataclass
class ProgressiveResizingHparams(AlgorithmHparams):
    """ Hyperparameters for the 'progressive resizing' algorithm
    """

    mode: str = hp.optional(doc="Type of scaling to perform", default="resize")
    initial_scale: float = hp.optional(doc="Initial scale factor", default=0.5)
    finetune_fraction: float = hp.optional(doc="Fraction of training to reserve for finetuning on full-sized inputs",
                                           default=0.2)
    resize_targets: bool = hp.optional(doc="Also resize targets", default=False)

    def initialize_object(self) -> ProgressiveResizing:
        return ProgressiveResizing(**asdict(self))


class ProgressiveResizing(Algorithm):
    """
    Applies the 'progressive resizing' data augmentation algorithm to speed up training.
    See `Training a State-of-the-Art Model` <https://github.com/fastai/fastbook/blob/780b76bef3127ce5b64f8230fce60e915a7e0735/07_sizing_and_tta.ipynb>`__.

    "Progressive resizing" initially scales inputs down to speed up early training.
    Throughout training, the scaling factor is gradually increased, yielding larger inputs
    up to the original input size. A final finetuning period is then run to finetune the
    model using the full-sized inputs.

    Args:
        mode (str): Type of scaling to perform. Value must be one of 'crop' or 'resize'.
                    'crop' performs a random crop, whereas 'resize' performs a bilinear
                    interpolation. Default: 'resize'.
        initial_scale(float): Initial scale factor used to shrink the inputs. Must be a
                              value in between 0 and 1.
        finetune_fraction (float): Fraction of training to reserve for finetuning on the
                                   full-sized inputs. Must be a value in between 0 and 1.
        resize_targets (bool): If True, resize targets also.
    """

    def __init__(self, mode: str, initial_scale: float, finetune_fraction: float, resize_targets: bool):

        if mode not in _VALID_MODES:
            raise ValueError(f"mode '{mode}' is not supported. Must be one of {_VALID_MODES}")

        if not (0 <= initial_scale <= 1):
            raise ValueError(f"initial_scale must be between 0 and 1: {initial_scale}")

        if not (0 <= finetune_fraction <= 1):
            raise ValueError(f"finetune_fraction must be between 0 and 1: {finetune_fraction}")

        self.hparams = ProgressiveResizingHparams(mode=mode,
                                                  initial_scale=initial_scale,
                                                  finetune_fraction=finetune_fraction,
                                                  resize_targets=resize_targets)

    def match(self, event: Event, state: State) -> bool:
        """ Apply on Event.AFTER_DATALOADER """
        return event == Event.AFTER_DATALOADER

    def apply(self, event: Event, state: State, logger: Optional[Logger] = None) -> None:
        input, target = state.batch_pair
        assert isinstance(input, Tensor) and isinstance(target, Tensor), \
            "Multiple tensors not supported for this method yet."

        # Calculate the current size of the inputs to use
        initial_size = self.hparams.initial_scale
        finetune_fraction = self.hparams.finetune_fraction
        scale_frac_elapsed = min([(state.epoch / state.max_epochs) / (1 - finetune_fraction), 1])

        # Linearly increase to full size at the start of the fine tuning period
        scale_factor = initial_size + (1 - initial_size) * scale_frac_elapsed

        new_input, new_target = resize_inputs(X=input,
                                              y=target,
                                              scale_factor=scale_factor,
                                              mode=self.hparams.mode,
                                              resize_targets=self.hparams.resize_targets)
        state.batch = (new_input, new_target)



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Iterator

import torch
import torch.distributed
import torch.utils.data
import yahp as hp
from torch.utils.data.sampler import Sampler

from composer.core.types import Batch, DataLoader
from composer.datasets.hparams import DataloaderSpec


class WrappedDataLoader(DataLoader):

    def __init__(self, dataloader: DataLoader) -> None:
        self.dataset = dataloader.dataset
        self.batch_size = dataloader.batch_size
        self.num_workers = dataloader.num_workers
        self.pin_memory = dataloader.pin_memory
        self.drop_last = dataloader.drop_last
        self.timeout = dataloader.timeout
        self.sampler = dataloader.sampler
        self.prefetch_factor = dataloader.prefetch_factor
        self.dataloader = dataloader

    def __len__(self) -> int:
        return len(self.dataloader)

    def __iter__(self) -> Iterator[Batch]:
        return iter(self.dataloader)

    def __bool__(self) -> bool:
        return True

    def __setattr__(self, name: str, value: Any) -> None:
        if hasattr(self, name) and name in ("dataset", "batch_size", "num_workers", "pin_memory", "drop_last",
                                            "timeout", "sampler", "prefetch_factor", "dataloader"):
            raise RuntimeError(f"Property {name} cannot be set after initialization in a DataLoader")
        return super().__setattr__(name, value)


@dataclass
class DataloaderHparams(hp.Hparams):
    num_workers: int = hp.required(doc="Number of CPU workers to use per gpu", template_default=8)
    prefetch_factor: int = hp.required(doc="Number of samples loaded in advance by each worker", template_default=2)
    persistent_workers: bool = hp.required(doc="Whether or not to shutdown workers after the dataset"
                                           " has been consumed once",
                                           template_default=True)
    pin_memory: bool = hp.required(doc="Whether or not to copy Tensors into CUDA pinned memory"
                                   " before returning them",
                                   template_default=True)
    timeout: int = hp.required(doc="Timeout value for collecting a batch from workers. 0 for no timeout.",
                               template_default=0)

    def initialize_object(
        self,
        batch_size: int,
        sampler: Sampler,
        dataloader_spec: DataloaderSpec,
    ) -> DataLoader:
        dataloader = torch.utils.data.DataLoader(
            dataloader_spec.dataset,
            batch_size=batch_size,
            shuffle=False,  # set in the sampler
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            drop_last=dataloader_spec.drop_last,
            sampler=sampler,
            collate_fn=dataloader_spec.collate_fn,
            worker_init_fn=dataloader_spec.worker_init_fn,
            multiprocessing_context=dataloader_spec.multiprocessing_context,
            generator=dataloader_spec.generator,
            timeout=self.timeout,
            prefetch_factor=self.prefetch_factor,
            persistent_workers=self.persistent_workers)
        return dataloader



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import logging
from dataclasses import asdict, dataclass
from typing import Optional

import torch
import yahp as hp

from composer.algorithms import AlgorithmHparams
from composer.core import Algorithm, Event, Logger, State
from composer.utils import ensure_tuple

log = logging.getLogger(__name__)


@dataclass
class SAMHparams(AlgorithmHparams):
    rho: float = hp.optional(doc='The neighborhood size parameter of SAM. Must be greater than 0.', default=0.05)
    epsilon: float = hp.optional(doc='A small value added to gradient norm for numerical stability.', default=1.0e-12)
    interval: int = hp.optional(doc='SAM will run once per `interval` steps. A value of 1 will cause'
                                'SAM to run every step. Steps on which SAM runs take roughly twice'
                                'as much time to complete.',
                                default=1)

    def initialize_object(self) -> SAM:
        return SAM(**asdict(self))


class SAMOptimizer(torch.optim.Optimizer):
    """Implementation based on https://github.com/davda54/sam"""

    def __init__(self, base_optimizer, rho, epsilon, interval, **kwargs):
        assert rho >= 0.0, f"Invalid rho, should be non-negative: {rho}"
        self.base_optimizer = base_optimizer
        self.global_step = 0
        self.interval = interval
        self._step_supports_amp_closure = True  # Flag for Mosaic trainer
        defaults = dict(rho=rho, epsilon=epsilon, **kwargs)
        super(SAMOptimizer, self).__init__(self.base_optimizer.param_groups, defaults)

    @torch.no_grad()
    def sub_e_w(self):
        for group in self.param_groups:
            for p in group["params"]:
                if "e_w" not in self.state[p]:
                    continue
                e_w = self.state[p]["e_w"]  # retrieve stale e(w)
                p.sub_(e_w)  # get back to "w" from "w + e(w)"

    @torch.no_grad()
    def first_step(self):
        grad_norm = self._grad_norm()
        for group in self.param_groups:
            scale = group["rho"] / (grad_norm + group["epsilon"])
            for p in group["params"]:
                if p.grad is None:
                    continue
                e_w = p.grad * scale.to(p)
                p.add_(e_w)  # climb to the local maximum "w + e(w)"
                self.state[p]["e_w"] = e_w

    @torch.no_grad()
    def second_step(self):
        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None or "e_w" not in self.state[p]:
                    continue
                p.sub_(self.state[p]["e_w"])  # get back to "w" from "w + e(w)"
        self.base_optimizer.step()  # do the actual "sharpness-aware" update

    @torch.no_grad()
    def step(self, closure=None):
        assert closure is not None, "Sharpness Aware Minimization requires closure, but it was not provided"
        closure = torch.enable_grad()(closure)  # the closure should do a full forward-backward pass
        loss = None

        if (self.global_step + 1) % self.interval == 0:
            loss = closure(ddp_sync=False)  # Compute gradient at (w) per-GPU, and do not sync
            if loss:
                self.first_step()  # Compute e(w) and set weights to (w + (e(w)) separately per-GPU
                if closure():  # Compute gradient at (w + e(w))
                    self.second_step()  # Reset weights to (w) and step base optimizer
                else:
                    self.sub_e_w()  # If second forward-backward closure fails, reset weights to (w)
        else:
            loss = closure()
            if loss:
                self.base_optimizer.step()

        self.global_step += 1
        return loss

    def _grad_norm(self):
        norm = torch.norm(torch.stack(
            [p.grad.norm(p=2) for group in self.param_groups for p in group["params"] if p.grad is not None]),
                          p=2)
        return norm


class SAM(Algorithm):
    """Applies SAM by wrapping existing optimizers with the SAMOptimizer."""

    def __init__(
        self,
        rho: float = 0.05,
        epsilon: float = 1.0e-12,
        interval: int = 1,
    ):
        """
        __init__ is constructed from the same fields as in hparams.
        """
        self.hparams = SAMHparams(rho=rho, epsilon=epsilon, interval=interval)

    def match(self, event: Event, state: State) -> bool:
        return event == Event.TRAINING_START

    def apply(self, event: Event, state: State, logger: Optional[Logger]) -> Optional[int]:
        assert state.optimizers is not None

        state.optimizers = tuple(
            SAMOptimizer(
                base_optimizer=optimizer,
                rho=self.hparams.rho,
                epsilon=self.hparams.epsilon,
                interval=self.hparams.interval,
            ) for optimizer in ensure_tuple(state.optimizers))



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Protocol, Sequence, Tuple, Union

import torch
import torch.utils.data
from torchmetrics.collections import MetricCollection
from torchmetrics.metric import Metric

from composer.core.algorithm import Algorithm
from composer.core.event import Event as Event
from composer.core.logging import Logger as Logger
from composer.core.precision import Precision as Precision
from composer.core.serializable import Serializable as Serializable
from composer.core.state import State as State

Tensor = torch.Tensor
Tensors = Union[Tensor, Tuple[Tensor, ...], List[Tensor]]
"""
The Mosaic Trainer supports multiple type of Batches returned
from the dataloader:
* (Tensors, Tensors): pair of tensors for the target and labels.
* Dict[str, Any]: dictionary of data, typically used in NLP.
"""
BatchPair = Sequence[Tensors]
BatchDict = Dict[str, Tensor]
Batch = Union[BatchPair, BatchDict, Tensor]


def as_batch_dict(batch: Batch) -> BatchDict:
    if not isinstance(batch, dict):
        raise TypeError(f'batch_dict requires batch of type dict, got {type(batch)}')
    return batch


def as_batch_pair(batch: Batch) -> BatchPair:
    if not isinstance(batch, (tuple, list)):
        raise TypeError(f'batch_pair required batch to be a tuple or list, got {type(batch)}')
    if not len(batch) == 2:
        raise TypeError(f'batch has length {len(batch)}, expected length 2')
    return batch


Dataset = torch.utils.data.Dataset[Batch]


class BreakEpochException(Exception):
    pass


class DataLoader(Protocol):
    dataset: Dataset
    batch_size: Optional[int]
    num_workers: int
    pin_memory: bool
    drop_last: bool
    timeout: float
    sampler: torch.utils.data.Sampler[int]
    prefetch_factor: int

    def __iter__(self) -> Iterator[Batch]:
        ...

    def __len__(self) -> int:
        ...


Metrics = Union[Metric, MetricCollection]

Optimizer = torch.optim.Optimizer
Optimizers = Union[Optimizer, Tuple[Optimizer, ...]]
Scheduler = torch.optim.lr_scheduler._LRScheduler
Schedulers = Union[Scheduler, Tuple[Scheduler, ...]]

Scaler = torch.cuda.amp.grad_scaler.GradScaler

Model = torch.nn.Module
ModelParameters = Union[Iterable[Tensor], Iterable[Dict[str, Tensor]]]

Algorithms = Sequence[Algorithm]

JSON = Union[str, float, int, None, List['JSON'], Dict[str, 'JSON']]

TPrefetchFn = Callable[[Batch], Batch]

StateDict = Dict[str, Any]



# Copyright 2021 MosaicML. All Rights Reserved.

"""
Helper functions for auto-generating tables from metadata
"""
import importlib
import os
import shutil
import tempfile


def list_dirs(folder):
    """
    Lists all dirs for a given folder
    """
    return sorted(
        child.name for child in folder.iterdir() if child.is_dir() and ("__pycache__" not in str(child.absolute())))


def assert_attributes_exist(module, attributes):
    """
    Assert that module has the provided attributes
    """

    for attribute in attributes:
        assert hasattr(module, attribute), \
        f"{module} should define {attribute} in its __init__.py file."


def get_metadata(names, attributes, module_basepath):
    """
    Returns a nested dict of metadata with names as keys. Checks
    that all attributes exist in module given by module_basepath.name.

    Example:
        >>> get_metadata(
                names=['blurpool', 'label_smoothing'],
                attributes=['_name', '_tldr'],
                module_basepath='composer.algorithms'
            )
        {'blurpool': {'_name': ..., '_tldr': ...},
         'label_smoothing': {'_name': ..., '_tldr': ...}}
    """
    metadata = {}

    for name in names:
        module = importlib.import_module(f'{module_basepath}.{name}')
        assert_attributes_exist(module, attributes)

        metadata[name] = {a: getattr(module, a) for a in attributes}

        # check for attributes with empty strings
        for attribute in attributes:
            if not metadata[name][attribute]:
                print(f'WARNING: {name} has empty metadata {attribute}')
    return metadata


def build_markdown_table(header, metadata, sorted_keys, row_format):
    """
    Builds a markdown table, formatting `row_format` with the `metadata`. Entries
    in the table are ordered by `sorted_keys`.

    Args:
        header (list): list of header strings
        metadata (dict): nested dict of metadata
        sorted_keys (list): order of rows in table
        row_format (list): list of length(header). Elements are either a string
                           or a single-argument callable that returns a string.

    Returns:
        table_md (list): table in markdown format
    """
    table_md = print_row(header)
    table_md += print_row(['-' * len(h) for h in header])

    for name in sorted_keys:
        d = metadata[name]

        # single-argument callable that returns a string is used for conditional formats
        # e.g. to only print link if provided, define
        #    lambda d: '[Link]({_link})' if d[_link] else ''
        row = [r(d).format(**d) if callable(r) else r.format(**d) for r in row_format]
        table_md += print_row(row)

    return table_md


def print_row(row):
    return '|'.join(row) + '\n'


def index_tag_in_lines(lines, tag):
    """
    Returns line number where tag is found.
    """
    for index, line in enumerate(lines):
        if tag in line:
            return index
    raise ValueError(f'{tag} not found.')


def update_table_in_file(table, source_file):
    """
    Given a `source file`, updates the table. Searches
    the file for 'Table Start' and 'Table End' tags, and replaces
    the content between those tags.

    The original file is retained with the `.bkp` suffix.

    Args:
        table (list): list of strings
        source_file (path): path to source file
    """
    with open(source_file, 'r') as source, \
      tempfile.NamedTemporaryFile('w', delete=False) as temp:
        source_lines = source.readlines()

        table_start = index_tag_in_lines(source_lines, tag='Table Start')
        table_end = index_tag_in_lines(source_lines, tag='Table End')
        print(f'Found table_start tag at line no: {table_start}')
        print(f'Found table_end tag at line no: {table_end}')
        assert table_end > table_start, "Table End must be after Table Start"

        table_written = False
        for line_no, line in enumerate(source_lines):
            if line_no <= table_start or line_no >= table_end:
                temp.write(line)
            elif not table_written:  # write table once
                temp.writelines(table)
                table_written = True

    backup_file = source_file.with_suffix('.md.bkp')
    os.rename(source_file, backup_file)
    print(f'Original file backed up at: {backup_file}')

    shutil.copy(temp.name, source_file)
    print(f'Updated table in {source_file} (now {os.path.getsize(source_file)} bytes)')



# Copyright 2021 MosaicML. All Rights Reserved.

from typing import List, Optional

from composer.models.base import MosaicClassifier
from composer.models.model_hparams import Initializer
from composer.models.resnets import ImageNet_ResNet


class ResNet50(MosaicClassifier):

    def __init__(
        self,
        num_classes: int,
        initializers: Optional[List[Initializer]] = None,
    ) -> None:
        if initializers is None:
            initializers = []

        model = ImageNet_ResNet.get_model_from_name(
            "imagenet_resnet_50",
            initializers,
            num_classes,
        )
        super().__init__(module=model)



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.core.logging.base_backend import BaseLoggerBackend as BaseLoggerBackend
from composer.core.logging.base_backend import RankZeroLoggerBackend as RankZeroLoggerBackend
from composer.core.logging.logger import Logger as Logger
from composer.core.logging.logger import LogLevel as LogLevel
from composer.core.logging.logger import TLogData as TLogData
from composer.core.logging.logger import TLogDataValue as TLogDataValue
from composer.loggers.logger_hparams import BaseLoggerBackendHparams as BaseLoggerBackendHparams
from composer.loggers.logger_hparams import FileLoggerBackendHparams as FileLoggerBackendHparams
from composer.loggers.logger_hparams import WandBLoggerBackendHparams as WandBLoggerBackendHparams
from composer.loggers.registry import logger_registry as logger_registry



# Copyright 2021 MosaicML. All Rights Reserved.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, List, Optional

import torch
import torch.nn as nn
import yahp as hp

from composer.models.base import BaseMosaicModel
from composer.utils.string_enum import StringEnum


class Initializer(StringEnum):
    KAIMING_NORMAL = "kaiming_normal"
    KAIMING_UNIFORM = "kaiming_uniform"
    BN_UNIFORM = "bn_uniform"
    BN_ONES = "bn_ones"
    XAVIER_UNIFORM = "xavier_uniform"

    def get_initializer(self) -> Callable[[torch.nn.Module], None]:

        def kaiming_normal(w: nn.Module):
            if isinstance(w, torch.nn.Linear) or isinstance(w, torch.nn.Conv2d):
                torch.nn.init.kaiming_normal_(w.weight)

        def kaiming_uniform(w: nn.Module):
            if isinstance(w, torch.nn.Linear) or isinstance(w, torch.nn.Conv2d):
                torch.nn.init.kaiming_uniform_(w.weight)

        def xavier_uniform(w: nn.Module):
            if isinstance(w, torch.nn.Linear) or isinstance(w, torch.nn.Conv2d):
                torch.nn.init.xavier_uniform_(w.weight)

        def bn_ones(w: nn.Module):
            if isinstance(w, torch.nn.BatchNorm2d):
                w.weight.data = torch.ones_like(w.weight.data.shape)
                w.bias.data = torch.zeros_like(w.bias.data)

        def bn_uniform(w: nn.Module):
            if isinstance(w, torch.nn.BatchNorm2d):
                w.weight.data = torch.rand(w.weight.data.shape)
                w.bias.data = torch.zeros_like(w.bias.data)

        initializer_dict = {
            "kaiming_normal": kaiming_normal,
            "kaiming_uniform": kaiming_uniform,
            "bn_uniform": bn_uniform,
            "bn_ones": bn_ones,
            "xavier_uniform": xavier_uniform
        }
        if self.value not in initializer_dict:
            raise ValueError(f"Initializer '{self.value}' not found.")
        return initializer_dict[self.value]


@dataclass
class ModelHparams(hp.Hparams, ABC):

    initializers: List[Initializer] = hp.optional(
        default_factory=lambda: [],
        doc="The initialization strategy for the model",
    )

    num_classes: Optional[int] = hp.optional(
        doc="The number of classes.  Needed for classification tasks",
        default=None,
    )

    @abstractmethod
    def initialize_object(self) -> BaseMosaicModel:
        pass



# Copyright 2021 MosaicML. All Rights Reserved.

import pytest

from composer.models import BaseMosaicModel, ModelHparams
from composer.trainer.trainer_hparams import model_registry


@pytest.mark.parametrize("model_name", model_registry.keys())
def test_model_registry(model_name, request):
    if model_name == 'gpt2':  # do not pull from HF model hub
        request.applymarker(pytest.mark.xfail())

    # create the model hparams object
    model_hparams = model_registry[model_name]()

    requires_num_classes = set([
        "resnet56_cifar10",
        "efficientnetb0",
        "resnet101",
        "resnet50",
        "resnet18",
        "mnist_classifier",
    ])
    if model_name in requires_num_classes:
        model_hparams.num_classes = 10

    assert isinstance(model_hparams, ModelHparams)

    # create the model object using the hparams
    model = model_hparams.initialize_object()
    assert isinstance(model, BaseMosaicModel)



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import atexit
import os
import sys
from typing import Any, Dict, Optional, TextIO

import yaml

from composer.core.logging import Logger, LogLevel, RankZeroLoggerBackend, TLogData, format_log_data_value
from composer.core.state import State
from composer.loggers.logger_hparams import FileLoggerBackendHparams


class FileLoggerBackend(RankZeroLoggerBackend):
    """Log to a file.

    It prints logs to a file
    (or terminal, defaults to stdout) as specified by :param filename:.
    The :param log_level: 
    (defaults to :attr:`~composore.core.logging.logger.LogLevel.EPOCH`)
    controls the verbosity of the logging.

    Example output:
        ```
        [FIT][step=2]: { "logged_metric": "logged_value", }
        [EPOCH][step=2]: { "logged_metric": "logged_value", }
        [BATCH][step=2]: { "logged_metric": "logged_value", }
        [EPOCH][step=3]: { "logged_metric": "logged_value", }
        ```


    Args:
        filename (str, optional): File to log to. Can be a filepath, 'stdout', or 'stderr'. Defaults to 'stdout'.
        buffer_size (int, optional): Buffer size. See
            https://docs.python.org/3/library/functions.html#open. Defaults to 1
            for line buffering.
        log_level (LogLevel, optional): Maximum
            :class:`~composore.core.logging.logger.LogLevel`. to record.
            Defaults to :attr:`~composore.core.logging.logger.LogLevel.EPOCH`.
        every_n_epochs (int, optional):
            Frequency to print :attr:`~composore.core.logging.logger.LogLevel.EPOCH` logs.
            Defaults to 1 (every epoch).
        every_n_batches (int, optional):
            Frequency to print :attr:`~composore.core.logging.logger.LogLevel.BATCH` logs.
            Defaults to 1 (every batch).
        flush_every_n_batches (int, optional): How frequently to flush the log to the file.
            Defaults to 1 (every batch).
    """

    def __init__(
        self,
        filename: str = 'stdout',
        *,
        buffer_size: int = 1,
        log_level: LogLevel = LogLevel.EPOCH,
        every_n_epochs: int = 1,
        every_n_batches: int = 1,
        flush_every_n_batches: int = 1,
        config: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__()
        self.hparams = FileLoggerBackendHparams(
            filename=filename,
            buffer_size=buffer_size,
            log_level=log_level,
            every_n_epochs=every_n_epochs,
            every_n_batches=every_n_batches,
            flush_every_n_batches=flush_every_n_batches,
        )
        self.file: Optional[TextIO] = None
        self.config = config

    def _will_log(self, state: State, log_level: LogLevel) -> bool:
        if log_level > self.hparams.log_level:
            return False
        if log_level >= LogLevel.EPOCH and state.epoch % self.hparams.every_n_epochs != 0:
            return False
        if log_level >= LogLevel.BATCH and (state.step + 1) % self.hparams.every_n_batches != 0:
            return False
        return True

    def _log_metric(self, epoch: int, step: int, log_level: LogLevel, data: TLogData):
        data_str = format_log_data_value(data)
        print(f"[{log_level.name}][step={step}]: {data_str}", file=self.file)

    def _training_start(self, state: State, logger: Logger) -> None:
        if self.hparams.filename == "stdout":
            self.file = sys.stdout
        elif self.hparams.filename == "stderr":
            self.file = sys.stderr
        else:
            self.file = open(self.hparams.filename, "x+", buffering=self.hparams.buffer_size)
            atexit.register(self._close_file)
        if self.config is not None:
            print("Config", file=self.file)
            print("-" * 30, file=self.file)
            yaml.safe_dump(self.config, stream=self.file)
            print("-" * 30, file=self.file)
            print(file=self.file)

    def batch_end(self, state: State, logger: Logger) -> None:
        assert self.file is not None
        if (state.step + 1) % self.hparams.flush_every_n_batches == 0 and self.file not in (sys.stdout, sys.stderr):
            self.file.flush()
            os.fsync(self.file.fileno())

    def _close_file(self) -> None:
        assert self.file is not None
        assert self.file not in (sys.stdout, sys.stderr)
        self.file.close()



# Copyright 2021 MosaicML. All Rights Reserved.

import functools

import numpy as np
import pytest
import torch
from PIL import Image

from composer.algorithms import ColOut, ColOutHparams
from composer.algorithms.colout.colout import ColOutTransform, batch_colout
from composer.core import Event
from composer.trainer import TrainerHparams
from tests.utils.trainer_fit import train_model


def verify_shape_image(orig: Image.Image, new: Image.Image, p_row: float, p_col: float) -> None:
    """ Verify the shape of a transformed PIL Image """
    H_o, W_o = orig.height, orig.width
    H_n, W_n = new.height, new.width

    H_t = int((1 - p_row) * H_o)
    W_t = int((1 - p_col) * W_o)

    assert (H_n, W_n) == (H_t, W_t), f"Image shape mismatch: {(H_n, W_n)} != {(H_t, W_t)}"


def verify_shape_tensor(orig: torch.Tensor, new: torch.Tensor, p_row: float, p_col: float) -> None:
    """ Verify the shape of a transformed image tensor """
    C, H_o, W_o = orig.shape

    H_t = int((1 - p_row) * H_o)
    W_t = int((1 - p_col) * W_o)

    assert new.shape == (C, H_t, W_t), f"Image tensor shape mismatch: {new.shape} != {(C, H_t, W_t)}"


def verify_shape_batch(orig: torch.Tensor, new: torch.Tensor, p_row: float, p_col: float) -> None:
    """ Verify the shape of a transformed batch of images """
    N, C, H_o, W_o = orig.shape

    H_t = int((1 - p_row) * H_o)
    W_t = int((1 - p_col) * W_o)

    assert new.shape == (N, C, H_t, W_t), f"Image batch shape mismatch: {new.shape} != {(N, C, H_t, W_t)}"


@pytest.fixture(params=[False, True])
def batch(request) -> bool:
    """ Algorithm batch parameter """
    return request.param


@pytest.fixture(params=[0, 0.15])
def p_row(request) -> float:
    """ Algorithm p_row parameter """
    return request.param


@pytest.fixture
def p_col(p_row) -> float:
    """ Algorithm p_col parameter """
    return p_row


@pytest.fixture(params=[1, 3])
def C(request) -> int:
    """ Number of image channels. Testing BW and RGB. """
    return request.param


@pytest.fixture
def H(request) -> int:
    """ Default image height """
    return 32


@pytest.fixture
def W(H) -> int:
    """ Default image width (equal to height) """
    return H


@pytest.fixture
def fake_image(H, W, C) -> Image.Image:
    """ Fake PIL Image """
    rng = np.random.RandomState(0)
    return Image.fromarray((255 * rng.uniform(size=(H, W, C)).squeeze()).astype(np.uint8))


@pytest.fixture
def fake_image_tensor(H, W, C) -> torch.Tensor:
    """ Fake image tensor """
    torch.manual_seed(0)
    return torch.rand(C, H, W)


@pytest.fixture
def fake_image_batch(H, W, C) -> torch.Tensor:
    """ Fake batch of images """
    torch.manual_seed(0)
    return torch.rand(16, C, H, W)


@pytest.fixture
def dummy_hparams(p_row, p_col, batch) -> ColOutHparams:
    """ Dummy hparams """
    return ColOutHparams(p_row, p_col, batch)


@pytest.fixture
def dummy_algorithm(dummy_hparams) -> ColOut:
    """ Reusable algorithm instance """
    return ColOut(dummy_hparams.p_row, dummy_hparams.p_col, dummy_hparams.batch)


def test_single_image_drop_size(fake_image, p_row, p_col):
    """ Test application to single PIL image """
    transform = ColOutTransform(p_row, p_col)
    new_image = transform(fake_image)
    verify_shape_image(fake_image, new_image, p_row, p_col)  # type: ignore


@pytest.mark.parametrize("W", [48])
def test_rectangular_image(fake_image, p_row, p_col):
    """ Test application to a rectangular PIL image """
    transform = ColOutTransform(p_row, p_col)
    new_image = transform(fake_image)
    verify_shape_image(fake_image, new_image, p_row, p_col)  # type: ignore


def test_single_image_tensor_drop_size(fake_image_tensor, p_row, p_col):
    """ Test application to a single torch image tensor """
    transform = ColOutTransform(p_row, p_col)
    new_image = transform(fake_image_tensor)
    verify_shape_tensor(fake_image_tensor, new_image, p_row, p_col)  # type: ignore


def test_reproducibility_image(fake_image_tensor, p_row, p_col):
    """ Test that transform is reproducible given the same seed """
    transform_1 = ColOutTransform(p_row, p_col)
    transform_2 = ColOutTransform(p_row, p_col)

    torch.manual_seed(42)
    new_image_1 = transform_1(fake_image_tensor)
    torch.manual_seed(42)
    new_image_2 = transform_2(fake_image_tensor)

    assert torch.allclose(new_image_1, new_image_2)  # type: ignore


def test_reproducibility_batch(fake_image_batch, p_row, p_col):
    """ Test that batch augmentation is reproducible given the same seed """
    transform_1 = functools.partial(batch_colout, p_row=p_row, p_col=p_col)
    transform_2 = functools.partial(batch_colout, p_row=p_row, p_col=p_col)

    torch.manual_seed(42)
    new_batch_1 = transform_1(fake_image_batch)
    torch.manual_seed(42)
    new_batch_2 = transform_2(fake_image_batch)

    assert torch.allclose(new_batch_1, new_batch_2)  # type: ignore


def test_batch_drop_size(fake_image_batch, p_row, p_col):
    """ Test application to a batch of images """
    colout = functools.partial(batch_colout, p_row=p_row, p_col=p_col)
    new_batch = colout(fake_image_batch)
    verify_shape_batch(fake_image_batch, new_batch, p_row, p_col)


@pytest.mark.parametrize("p_col", [0.05, 0.25])
def test_rectangle_batch_drop_size(fake_image_batch, p_row, p_col):
    """ Test that unequal values of p_row and p_col work properly """
    colout = functools.partial(batch_colout, p_row=p_row, p_col=p_col)
    new_batch = colout(fake_image_batch)
    verify_shape_batch(fake_image_batch, new_batch, p_row, p_col)


@pytest.mark.parametrize("p_row,p_col", [(1.5, 0.15), (0.15, 1.5)])
def test_invalid_hparams(p_row, p_col):
    """ Test that invalid hyperparameters error.
    Ideally this could be caught by the Hparams, but that's not yet supported in yahp.
    """
    with pytest.raises(ValueError):
        ColOut(p_row, p_col, False)


@pytest.mark.parametrize("event,batch", [(Event.AFTER_DATALOADER, True), (Event.TRAINING_START, False)])
def test_match_correct(event, dummy_algorithm, dummy_state):
    """ Algo should match AFTER_DATALOADER if batch else TRAINING_START """
    assert dummy_algorithm.match(event, dummy_state)


@pytest.mark.parametrize("event,batch", [(Event.TRAINING_START, True), (Event.AFTER_DATALOADER, False),
                                         (Event.EPOCH_END, True)])
def test_match_incorrect(event, dummy_algorithm, dummy_state):
    """ Algo should NOT match TRAINING_START if batch else AFTER_DATALOADER """
    assert not dummy_algorithm.match(event, dummy_state)


@pytest.mark.parametrize("batch", [True])
def test_apply_batch(fake_image_batch, dummy_algorithm, dummy_state, dummy_logger):
    """ Apply the algorithm to a fake batch """
    p_row = dummy_algorithm.hparams.p_row
    p_col = dummy_algorithm.hparams.p_col

    dummy_state.batch = (fake_image_batch, None)
    dummy_algorithm.apply(Event.AFTER_DATALOADER, dummy_state, dummy_logger)
    last_input, _ = dummy_state.batch
    verify_shape_batch(fake_image_batch, last_input, p_row, p_col)


@pytest.mark.parametrize("batch", [False])
def test_apply_sample(dummy_algorithm, dummy_state, dummy_train_dataloader, dummy_logger):
    """ Test that augmentation is added to dataset and functioning properly """
    p_row = dummy_algorithm.hparams.p_row
    p_col = dummy_algorithm.hparams.p_col

    dset = dummy_train_dataloader.dataset
    orig, _ = dset[0]

    dummy_state.train_dataloader = dummy_train_dataloader
    dummy_algorithm.apply(Event.TRAINING_START, dummy_state, dummy_logger)
    new, _ = dset[0]

    verify_shape_tensor(orig, new, p_row, p_col)


@pytest.mark.run_long
@pytest.mark.timeout(90)
def test_colout_trains(mosaic_trainer_hparams: TrainerHparams):
    mosaic_trainer_hparams.algorithms = [ColOutHparams(p_row=0.15, p_col=0.15, batch=True)]
    train_model(mosaic_trainer_hparams)



# Copyright 2021 MosaicML. All Rights Reserved.

import os

import pytest
import torch
import torch.distributed

import composer
from composer.core.types import Precision
from composer.trainer import TrainerHparams
from composer.trainer.ddp import FileStoreHparams
from composer.trainer.devices import GPUDeviceHparams


def run_and_measure_memory(precision: Precision, file_store_path: str) -> int:
    hparams_f = os.path.join(os.path.dirname(composer.__file__), "yamls", "models", "resnet56_cifar10",
                             "hparams_synthetic.yaml")
    hparams = TrainerHparams.create(f=hparams_f)
    assert isinstance(hparams, TrainerHparams)
    assert isinstance(hparams.device, GPUDeviceHparams)
    hparams.device.n_gpus = 1
    hparams.precision = precision
    hparams.ddp.store = FileStoreHparams(file_store_path)
    hparams.ddp.fork_rank_0 = False
    hparams.dataloader.num_workers = 0
    hparams.dataloader.persistent_workers = False
    hparams.max_epochs = 2
    hparams.loggers = []
    trainer = hparams.initialize_object()
    torch.cuda.empty_cache()
    torch.cuda.reset_peak_memory_stats()
    trainer.fit()
    return torch.cuda.max_memory_allocated()


@pytest.mark.timeout(60)
@pytest.mark.run_long
@pytest.mark.n_gpus(1)
def test_fp16_mixed(ddp_tmpdir: str):
    memory_full = run_and_measure_memory(Precision.FP32, os.path.join(ddp_tmpdir, "store_full"))
    memory_amp = run_and_measure_memory(Precision.AMP, os.path.join(ddp_tmpdir, "store_amp"))
    assert memory_amp < 0.7 * memory_full



import logging

import composer
from composer.trainer.trainer import Trainer
from composer.trainer.trainer_hparams import TrainerHparams


def main() -> None:
    logging.basicConfig()
    logging.captureWarnings(True)

    hparams = TrainerHparams.create()
    logging.getLogger(composer.__name__).setLevel(hparams.log_level)
    trainer = Trainer.create_from_hparams(hparams=hparams)
    trainer.fit()


if __name__ == "__main__":
    main()



# Copyright 2021 MosaicML. All Rights Reserved.

from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Generator, Optional, TypeVar, Union

import torch.nn

from composer.core.serializable import Serializable
from composer.core.state import State
from composer.core.types import DataLoader, Optimizer, Precision, Tensor, TPrefetchFn

T_nnModule = TypeVar("T_nnModule", bound=torch.nn.Module)


class Device(Serializable, ABC):

    @abstractmethod
    def prepare(self, state: State) -> None:
        """prepare() is invoked by the trainer at the beginning of the training loop.
        The device should perform any initialization here.
        It should not modify the state.

        Args:
            state (State): The global state variable.
        """

    @abstractmethod
    def module_to_device(self, module: T_nnModule) -> T_nnModule:
        """module_to_device moves a module onto the device's device.

        Args:
            module (T_nnModule): The module to move to the device

        Returns:
            T_nnModule: The module on the device.
        """
        pass

    @abstractmethod
    def tensor_to_device(self, tensor: Tensor) -> Tensor:
        """tensor_to_device moves a tensor onto the device's device.

        Args:
            tensor (T_nnModule): The tensor to move to the device

        Returns:
            Tensor: The tensor on the device.
        """
        pass

    @abstractmethod
    def dataloader_to_device(self, dataloader: DataLoader, prefetch_fn: Optional[TPrefetchFn]) -> DataLoader:
        """dataloader_to_device wraps a dataloader and ensures that all returned batches are on the
        the correct device. It is responsible for executing `prefetch_fn`, if provided, on each batch before it is
        yielded. The `prefetch_fn` can be executed in the background, if the device supports it.

        Args:
            dataloader (DataLoader): The dataloader to wrap.
            prefetch_fn (Optional[TPrefetchFn]): A function that takes a batch and returns a batch.
                It should perform any on-device preprocessing of a batch.
                (e.g. on a GPU device, this function can be used for gpu transformations.)

        Returns:
            DataLoader: The wrapped dataloader, which yields batches that have been moved to the device
                and have been processed through the prefetch_fn.
        """

    def optimizer_to_device(self, optimizer: Optimizer) -> Optimizer:
        """optimizer_to_device moves an optimizer's state onto the device's device.

        As a rule, this usually isn't necessary, since most optimizers lazy initialize their state
        when `.step()` is first called, based off of the device of the parameters. The prominent
        exception to this rule is when we are restoring from a checkpoint.

        Args:
            optimizer (Optimizer): The optimizer to move to the device

        Returns:
            Optimizer: The optimizer on the device
        """
        for state in optimizer.state.values():
            for k, v in state.items():
                if isinstance(v, Tensor):
                    state[k] = self.tensor_to_device(v)
        return optimizer

    @abstractmethod
    @contextmanager
    def precision_context(self, precision: Union[str, Precision]) -> Generator[None, None, None]:
        """precision returns a context manager that uses the specified precision.

        Example usage:

        ```
        with device.precision(Precision.AMP):
            forward_pass_with_amp()
        ```

        Args:
            precision (Precision): [description]

        Yields:
            Generator[None, None, None]: [description]
        """
        pass

    @property
    @abstractmethod
    def nproc_per_node(self) -> int:
        """nproc_per_node returns the number of processes per node the device would like to have.
        This function may be invoked before `prepare()` and therefore should not rely on state.

        Returns:
            int: Number of processes per node.
        """

    @property
    @abstractmethod
    def ddp_backend(self) -> str:
        """DDP backend to use. Should return `gloo`, `mpi`, or `nccl`.
        See https://pytorch.org/docs/stable/distributed.html

        Returns:
            str: `gloo`, `mpi`, or `nccl`
        """



# Copyright 2021 MosaicML. All Rights Reserved.

from unittest.mock import Mock

import pytest

from composer.core import Event
from composer.core.callback import Callback
from composer.core.engine import Engine
from composer.core.logging import Logger
from composer.core.state import State


def test_callbacks_map_to_events():
    # callback methods must be 1:1 mapping with events
    # exception for private methods
    cb = Callback()
    excluded_methods = ["state_dict", "load_state_dict", "setup"]
    methods = set(m for m in dir(cb) if (m not in excluded_methods and not m.startswith("_")))
    event_names = set(e.value for e in Event)
    assert methods == event_names


@pytest.mark.parametrize('event', list(Event))
def test_run_event_callbacks(event: Event, dummy_state: State):
    callbacks = [Mock() for _ in range(5)]
    logger = Logger(dummy_state)
    engine = Engine(state=dummy_state, algorithms=[], logger=logger, callbacks=callbacks)

    engine.run_event(event)

    for cb in callbacks:
        f = getattr(cb, event.value)
        f.assert_called_once_with(dummy_state, logger)



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.blurpool.blurpool import BlurPool as BlurPool
from composer.algorithms.blurpool.blurpool import BlurPoolHparams as BlurPoolHparams
from composer.algorithms.blurpool.blurpool import apply_blurpool as apply_blurpool
from composer.algorithms.blurpool.blurpool_layers import BlurConv2d as BlurConv2d
from composer.algorithms.blurpool.blurpool_layers import BlurMaxPool2d as BlurMaxPool2d
from composer.algorithms.blurpool.blurpool_layers import BlurPool2d as BlurPool2d
from composer.algorithms.blurpool.blurpool_layers import blur_2d as blur_2d

_name = 'BlurPool'
_class_name = 'BlurPool'
_functional = 'apply_blurpool'
_tldr = ''
_attribution = '(Zhang, 2019)'
_link = 'https://arxiv.org/abs/1904.11486'
_method_card = ''



# Copyright 2021 MosaicML. All Rights Reserved.

import atexit
from typing import Callable, List

import _pytest.config
import _pytest.config.argparsing
import _pytest.fixtures
import _pytest.mark
import pytest
from _pytest.monkeypatch import MonkeyPatch

N_GPU_OPTIONS = (0, 1, 2, 4, 8)

# Add the path of any pytest fixture files you want to make global
pytest_plugins = [
    "tests.fixtures.dummy_fixtures",
    "tests.fixtures.ddp_fixtures",
    "tests.algorithms.algorithm_fixtures",
]


def pytest_addoption(parser: _pytest.config.argparsing.Parser) -> None:
    parser.addoption(
        "--n_gpus",
        default=None,
        type=int,
        choices=N_GPU_OPTIONS,
        help=
        "Number of GPUs available. Only tests marked with @pytest.mark.n_gpus(val) will be executed. Defaults to all tests."
    )


def _get_n_gpus(item: pytest.Item) -> int:
    n_gpu_marks = tuple(item.iter_markers(name="n_gpus"))
    assert len(n_gpu_marks) <= 1
    if len(n_gpu_marks) == 1:
        n_gpu_mark = n_gpu_marks[0]
        return n_gpu_mark.args[0]
    else:
        return 0


def pytest_collection_modifyitems(session: pytest.Session, config: _pytest.config.Config,
                                  items: List[pytest.Item]) -> None:
    items_to_remove: List[pytest.Item] = []
    cli_n_gpus = config.getoption("n_gpus", default=None)
    if cli_n_gpus is None:
        return
    for item in items:
        test_n_gpus = _get_n_gpus(item)
        if cli_n_gpus != test_n_gpus:
            items_to_remove.append(item)
        if test_n_gpus not in N_GPU_OPTIONS:
            name = item.name
            raise ValueError(
                f"Invalid option @pytest.mark.n_gpus({test_n_gpus}) for test `{name}`: @pytest.mark.n_gpus() must be set to one of {N_GPU_OPTIONS}"
            )
    for item in items_to_remove:
        items.remove(item)


def parameterize_n_gpus(*n_gpus: int, field_name: str = "n_gpus") -> Callable[..., Callable[..., None]]:

    def wrapped_func(testfunc: Callable[..., None]) -> Callable[..., None]:
        return pytest.mark.parametrize(
            field_name, [pytest.param(n_gpu, marks=[pytest.mark.n_gpus(n_gpu)]) for n_gpu in n_gpus])(testfunc)

    return wrapped_func


@pytest.fixture(autouse=True)
def atexit_at_test_end(monkeypatch: MonkeyPatch):
    # monkeypatch atexit so it is called when a test exits, not when the python process exits
    atexit_callbacks = []

    def register(func, *args, **kwargs):
        atexit_callbacks.append((func, args, kwargs))

    monkeypatch.setattr(atexit, "register", register)
    yield
    for func, args, kwargs in atexit_callbacks:
        func(*args, **kwargs)



# Copyright 2021 MosaicML. All Rights Reserved.

from dataclasses import asdict, dataclass

import numpy as np
import torch
import yahp as hp
from PIL.Image import Image as ImageType

from composer.algorithms.algorithm_hparams import AlgorithmHparams
from composer.core.types import Algorithm, Event, List, Logger, State
from composer.utils.augmentation_primitives import augmentation_sets
from composer.utils.data import add_dataset_transform


@dataclass
class RandAugmentHparams(AlgorithmHparams):

    severity: int = hp.optional(doc="Intensity of each augmentation. Ranges from 0 (none) to 10 (maximum)", default=9)
    depth: int = hp.optional(doc="Number of augmentations to compose in a row", default=2)
    augmentation_set: str = hp.optional(
        doc=
        "Set of augmentations to sample from. 'all', 'safe' (only augmentations that don't appear on CIFAR10C/ImageNet10C), or 'original'",
        default="all")

    def initialize_object(self) -> "RandAugment":
        return RandAugment(**asdict(self))


def randaugment(img: ImageType = None,
                severity: int = 9,
                depth: int = 2,
                augmentation_set: List = augmentation_sets["all"]) -> ImageType:
    """
    Perform augmentations.
    """

    # Iterate over augmentations
    for _ in range(depth):
        aug = np.random.choice(augmentation_set)
        img = aug(img, severity)
    assert img is not None
    return img


class RandAugmentTransform(torch.nn.Module):

    def __init__(self, severity: int = 9, depth: int = 2, augmentation_set: str = "all"):
        """
        See documentation for RandAugment
        """
        super().__init__()
        if severity < 0 or severity > 10:
            raise ValueError("RandAugment severity value must be 0 ≤ severity ≤ 10")
        if depth < 0:
            raise ValueError("RandAugment depth value must be ≥ 0")
        if augmentation_set not in augmentation_sets.keys():
            raise KeyError(f"RandAugment augmentation_set is not one of {augmentation_sets.keys()}")
        self.severity = severity
        self.depth = depth
        self.augmentation_set = augmentation_sets[augmentation_set]

    def forward(self, img: ImageType) -> ImageType:

        return randaugment(img=img, severity=self.severity, depth=self.depth, augmentation_set=self.augmentation_set)


class RandAugment(Algorithm):
    """
    Object that does RandAugment (Cubuk et al. (2019), RandAugment: Practical
        automated data augmentation with a reduced search space). Can be passed as a
        transform to torchvision.transforms.Compose().

    Args:
        severity (int): Severity of augmentation operators (between 1 to 10). M in the
            original paper. Default = 9.
        depth (int): Depth of augmentation chain. N in the original paper Default = 2.
        augmentation_set (str): One of ["augmentations_all",
            "augmentations_corruption_safe", "augmentations_original"]. Set of
            augmentations to use. "augmentations_corruption_safe" excludes transforms
            that are part of the ImageNet-C/CIFAR10-C test sets.
            "augmentations_original" uses all augmentations, but some of the
            implementations are identical to the original github repo, which appears
            to contain implementation specificities for the augmentations "color",
            "contrast", "sharpness", and "brightness".
    """

    def __init__(self, severity: int = 9, depth: int = 2, augmentation_set: str = "all"):
        if severity < 0 or severity > 10:
            raise ValueError("RandAugment severity value must be 0 ≤ severity ≤ 10")
        if augmentation_set not in augmentation_sets.keys():
            raise KeyError(f"randaugment_augmentation_set is not one of {augmentation_sets.keys()}")
        self.hparams = RandAugmentHparams(severity=severity, depth=depth, augmentation_set=augmentation_set)

    def match(self, event: Event, state: State) -> bool:
        """ Runs on Event.TRAINING_START
        """
        return event == Event.TRAINING_START

    def apply(self, event: Event, state: State, logger: Logger) -> None:
        """ Inserts RandAugment into the list of dataloader transforms
        """
        ra = RandAugmentTransform(**self.hparams.to_dict())
        assert state.train_dataloader is not None
        dataset = state.train_dataloader.dataset
        add_dataset_transform(dataset, ra)



# Copyright 2021 MosaicML. All Rights Reserved.

from typing import Any

import numpy as np
import torch


def deep_compare(a: Any, b: Any) -> None:
    if isinstance(a, (list, tuple)):
        assert isinstance(b, type(a))
        assert len(a) == len(b)
        for x, y in zip(a, b):
            deep_compare(x, y)
        return
    elif isinstance(a, dict):
        assert isinstance(b, dict)
        assert len(a) == len(b)
        for k, v in a.items():
            deep_compare(v, b[k])
        return
    elif isinstance(a, (torch.Tensor, np.ndarray)):
        assert isinstance(b, type(a))
        assert a.shape == b.shape
        assert a.dtype == b.dtype
        assert (a == b).all()
        return
    else:
        assert a == b



# Copyright 2021 MosaicML. All Rights Reserved.

from typing import List, Optional, Sequence, Union

import torch
import torch.nn as nn
from torch.nn import functional as F

from composer.models.base import MosaicClassifier
from composer.models.model_hparams import Initializer


class Model(nn.Module):
    """Toy classifier for MNIST. Should not be used to evaluate any method."""

    def __init__(self, initializers: Sequence[Union[str, Initializer]], outputs: int):
        super().__init__()

        # TODO(issue #249):
        self.num_classes = outputs

        for initializer in initializers:
            initializer = Initializer(initializer)
            self.apply(initializer.get_initializer())

        self.conv1 = nn.Conv2d(1, 16, (3, 3), padding=0)
        self.conv2 = nn.Conv2d(16, 32, (3, 3), padding=0)
        self.bn = nn.BatchNorm2d(32)
        self.fc1 = nn.Linear(32 * 16, 32)
        self.fc2 = nn.Linear(32, outputs)

    def forward(self, x):
        out = self.conv1(x)
        out = F.relu(out)
        out = self.conv2(out)
        out = self.bn(out)
        out = F.relu(out)
        out = F.adaptive_avg_pool2d(out, (4, 4))
        out = torch.flatten(out, 1, -1)
        out = self.fc1(out)
        out = F.relu(out)
        return self.fc2(out)


class MNIST_Classifier(MosaicClassifier):

    def __init__(
        self,
        num_classes: int,
        initializers: Optional[List[Initializer]] = None,
    ) -> None:

        if initializers is None:
            initializers = []

        model = Model(initializers, num_classes)
        super().__init__(module=model)



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.layer_freezing.layer_freezing import LayerFreezing as LayerFreezing
from composer.algorithms.layer_freezing.layer_freezing import LayerFreezingHparams as LayerFreezingHparams
from composer.algorithms.layer_freezing.layer_freezing import freeze_layers as freeze_layers

_name = 'Layer Freezing'
_class_name = 'LayerFreezing'
_functional = 'freeze_layers'
_tldr = 'Progressively freezes layers during training.'
_attribution = 'Many (Raghu et al, 2017)'
_link = 'https://arxiv.org/abs/1706.05806'
_method_card = ''



# Copyright 2021 MosaicML. All Rights Reserved.

from dataclasses import asdict, dataclass
from typing import Optional

import yahp as hp

from composer.algorithms.algorithm_hparams import AlgorithmHparams
from composer.core.types import Algorithm, Event, Logger, State, Tensor
from composer.models.loss import ensure_targets_one_hot


@dataclass
class LabelSmoothingHparams(AlgorithmHparams):

    alpha: float = hp.required(doc='smoothing factor', template_default=0.1)

    def initialize_object(self) -> "LabelSmoothing":
        return LabelSmoothing(**asdict(self))


class LabelSmoothing(Algorithm):
    """Applies label smoothing during before_loss, then restores the
       original labels during after_loss.

       Args:
        alpha (float): Strength of the label smoothing, between [0, 1].
            alpha=0 means no label smoothing, and alpha=1 means maximal
            smoothing (targets are ignored)
    """

    def __init__(self, alpha: float):
        self.hparams = LabelSmoothingHparams(alpha=alpha)

    def match(self, event: Event, state: State) -> bool:
        return event in [Event.BEFORE_LOSS, Event.AFTER_LOSS]

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        input, labels = state.batch_pair

        if event == Event.BEFORE_LOSS:
            assert isinstance(state.outputs, Tensor), "Multiple tensors not supported yet"
            assert isinstance(labels, Tensor), "Multiple tensors not supported yet"

            self.original_labels = labels.clone()
            smoothed_labels = smooth_labels(
                state.outputs,
                labels,
                alpha=self.hparams.alpha,
            )
            state.batch = (input, smoothed_labels)
        elif event == Event.AFTER_LOSS:
            # restore the target to the non-smoothed version
            state.batch = (input, self.original_labels)


def smooth_labels(logits: Tensor, targets: Tensor, alpha: float):
    """Shrinks targets towards a prior distribution to counteract label noise.

    This is computed by `(1 - alpha) * targets + alpha * smoothed_targets`
    where `smoothed_targets` is a pre-specified vector of class probabilities.

    Introduced in: https://arxiv.org/abs/1512.00567
    Evaluated in: https://arxiv.org/abs/1906.02629

    Args:
        logits: Output of the model. Tensor of shape (N, C, d1, ..., dn) for
            N examples and C classes, and d1, ..., dn extra dimensions.
        targets: Tensor of shape (N) containing integers 0 <= i <= C-1
            specifying the target labels for each example.
        alpha: Strength of the label smoothing, between [0, 1].
            alpha=0 means no label smoothing, and alpha=1 means maximal
            smoothing (targets are ignored)
    """

    targets = ensure_targets_one_hot(logits, targets)
    n_classes = logits.shape[1]
    return (targets * (1. - alpha)) + (alpha / n_classes)



# Copyright 2021 MosaicML. All Rights Reserved.

import collections.abc
import os
from dataclasses import dataclass
from typing import Dict, List, Optional, Sequence
from unittest import mock

import pytest
import torch
import torch.distributed
import yahp as hp
from _pytest.monkeypatch import MonkeyPatch

import composer.core.types as types
from composer import Callback
from composer.callbacks import CallbackHparams
from composer.core.logging import Logger
from composer.core.state import State
from composer.datasets import DataloaderHparams, DataloaderSpec, MemoryFormat, SyntheticDataset, SyntheticDatasetHparams
from composer.models.classify_mnist import MnistClassifierHparams
from composer.trainer.ddp import DDPHparams, FileStoreHparams
from composer.trainer.devices import CPUDeviceHparams, GPUDeviceHparams
from composer.trainer.trainer_hparams import TrainerHparams, callback_registry, dataset_registry


def get_file_path(tmpdir: str, *, idx: int, epoch: int, is_train: bool) -> str:
    train_str = "train" if is_train else "val"
    return os.path.join(tmpdir, f"{train_str}-epoch-{epoch}-sample-{idx}")


def get_batch_file_path(tmpdir: str, *, rank: int, epoch: int, is_train: bool) -> str:
    train_str = "train" if is_train else "val"
    return os.path.join(tmpdir, f"{train_str}-rank-{rank}-epoch-{epoch}-batch0.pt")


class TrackedDataset(SyntheticDataset):
    """
    TrackedDataset atomically writes a file every time a record is accessed.
    It is thread-safe and subprocess-safe, and is useful to measure how many times a sample is accessed.
    Because of atomic file writes, it is slow and should not be used in any performance measurements.
    """

    def __init__(self, *, sample_pool_size: int, shape: Sequence[int], memory_format: MemoryFormat, device: str,
                 one_hot: bool, num_classes: int, is_train: bool, tmpdir: str):
        super().__init__(sample_pool_size=sample_pool_size,
                         shape=shape,
                         memory_format=memory_format,
                         device=device,
                         one_hot=one_hot,
                         num_classes=num_classes)
        self.is_train = is_train
        self.tmpdir = tmpdir

    def __getitem__(self, idx: int):
        access = 0
        while True:
            try:
                with open(get_file_path(self.tmpdir, idx=idx, epoch=access, is_train=self.is_train), "x") as f:
                    f.write(str(idx))
                return super().__getitem__(idx)
            except FileExistsError:
                access += 1


@dataclass
class TrackedDatasetHparams(SyntheticDatasetHparams):
    is_train: Optional[bool] = hp.optional("is_train", default=None)
    tmpdir: Optional[str] = hp.optional("tmpdir", default=None)

    def initialize_object(self) -> DataloaderSpec:
        assert self.is_train is not None
        assert self.tmpdir is not None
        return DataloaderSpec(
            TrackedDataset(
                num_classes=self.num_classes,
                shape=self.shape,
                one_hot=self.one_hot,
                device=self.device,
                memory_format=self.memory_format,
                sample_pool_size=self.sample_pool_size,
                is_train=self.is_train,
                tmpdir=self.tmpdir,
            ),
            drop_last=self.drop_last,
            shuffle=self.shuffle,
        )


class CheckBatch0(Callback):

    def __init__(self, tmpdir: str):
        super().__init__()
        self.tmpdir = tmpdir

    def before_forward(self, state: State, logger: Logger):
        if state.batch_idx > 0:
            return
        rank: int = torch.distributed.get_rank()
        last_input, last_target = state.batch_pair
        torch.save(  # type: ignore
            {
                "last_input": last_input,
                "last_target": last_target,
            }, get_batch_file_path(self.tmpdir, rank=rank, epoch=state.epoch, is_train=True))

    def eval_before_forward(self, state: State, logger: Logger):
        rank: int = torch.distributed.get_rank()
        filepath = get_batch_file_path(self.tmpdir, rank=rank, epoch=state.epoch, is_train=False)
        if os.path.exists(filepath):
            return
        assert not state.model.training
        last_input, last_target = state.batch_pair
        torch.save(  # type: ignore
            {
                "last_input": last_input,
                "last_target": last_target,
            }, get_batch_file_path(self.tmpdir, rank=rank, epoch=state.epoch, is_train=False))


@dataclass
class CheckBatch0Hparams(CallbackHparams):
    tmpdir: str = hp.required("tmpdir")

    def initialize_object(self) -> Callback:
        return CheckBatch0(self.tmpdir)


@pytest.fixture(autouse=True)
def patch_registries(monkeypatch: MonkeyPatch):
    monkeypatch.setitem(callback_registry, "checkbatch0", CheckBatch0Hparams)
    monkeypatch.setitem(dataset_registry, "tracked", TrackedDatasetHparams)


@pytest.mark.run_long
@pytest.mark.timeout(90)
@pytest.mark.parametrize("fork_rank_0", [True, False], ids=["fork-rank-0", "no-fork-rank-0"])
@pytest.mark.parametrize("is_gpu,num_procs", [
    pytest.param(False, 1, id="1-cpu"),
    pytest.param(False, 2, id="2-cpu"),
    pytest.param(True, 1, marks=[pytest.mark.n_gpus(1)], id="1-gpu"),
    pytest.param(True, 2, marks=[pytest.mark.n_gpus(2)], id="2-gpu"),
])
def test_ddp(is_gpu: bool, num_procs: int, fork_rank_0: bool, *, ddp_tmpdir: str, is_main_pytest_process: bool,
             mosaic_trainer_hparams: TrainerHparams) -> None:
    """
    test strategy for ddp:
    1) Train a dummy model on two gps, for two epochs, using the tracked dataset.
    2) The tracked dataset should record two -- and only two -- accesses for each sample -- one for each epoch
       If each sample is accessed more than this number of times, then the distributed sampler isn't working properly
       If each sample is accessed less than this number of times, then either the sample pool size isn't a multiple of
       the batch size (and samples are getting dropped), or not all processes are working
    3) We use a callback to save the (x, y) for the first batch in each epoch on each process
        ({train, eval} * {epoch 1, epoch 2} * {ddp 1, ddp2})
       We assert that each of these tensors are different to ensure that 1) random seeding works properly,
       and 2) each ddp process is indeed getting different data.
    """
    hparams = mosaic_trainer_hparams
    assert isinstance(hparams, TrainerHparams)
    hparams.model = MnistClassifierHparams(num_classes=10)
    mosaic_trainer_hparams.train_dataset = TrackedDatasetHparams(
        num_classes=10,
        shape=[1, 28, 28],
        sample_pool_size=300,
        one_hot=False,
        device="cpu",
        is_train=True,
        memory_format=MemoryFormat.CONTIGUOUS_FORMAT,
        tmpdir=ddp_tmpdir,
    )
    hparams.val_dataset = TrackedDatasetHparams(
        num_classes=10,
        shape=[1, 28, 28],
        sample_pool_size=300,
        one_hot=False,
        device="cpu",
        is_train=False,
        memory_format=MemoryFormat.CONTIGUOUS_FORMAT,
        tmpdir=ddp_tmpdir,
    )
    if is_gpu:
        device = GPUDeviceHparams(n_gpus=num_procs)
    else:
        device = CPUDeviceHparams(n_cpus=num_procs)
    hparams.device = device
    hparams.ddp = DDPHparams(
        store=FileStoreHparams(os.path.join(ddp_tmpdir, "store")),
        node_rank=0,
        num_nodes=1,
        fork_rank_0=fork_rank_0,
    )
    hparams.dataloader = DataloaderHparams(
        num_workers=0,
        prefetch_factor=2,
        persistent_workers=False,
        pin_memory=False,
        timeout=0,
    )
    hparams.total_batch_size = 50
    hparams.eval_batch_size = 50
    hparams.max_epochs = 2
    hparams.precision = types.Precision.FP32
    hparams.loggers = []
    hparams.validate_every_n_batches = 0
    hparams.validate_every_n_epochs = 1
    hparams.callbacks.append(CheckBatch0Hparams(tmpdir=ddp_tmpdir))
    trainer = hparams.initialize_object()
    assert trainer.state.world_size == num_procs
    assert trainer.state.nproc_per_node == num_procs
    assert isinstance(trainer.train_dl_spec.dataset, collections.abc.Sized)
    num_train_samples = len(trainer.train_dl_spec.dataset)
    assert isinstance(trainer.eval_dl_spec.dataset, collections.abc.Sized)
    num_eval_samples = len(trainer.eval_dl_spec.dataset)
    trainer.fit()

    # we want to validate on the spawning process only
    if is_main_pytest_process:
        # now validate that each sample were accessed exactly hparams.max_epochs * batch size times
        num_epochs = hparams.max_epochs

        for i in range(num_train_samples):
            for epoch in range(num_epochs):
                assert os.path.exists(
                    get_file_path(ddp_tmpdir, idx=i, epoch=epoch,
                                  is_train=True)), f"train sample {i} was not accessed during epoch {epoch}"
            assert not os.path.exists(get_file_path(ddp_tmpdir, idx=i, epoch=num_epochs,
                                                    is_train=True)), f"train sample {i} was accessed too many times"

        for i in range(num_eval_samples):
            for epoch in range(num_epochs):
                assert os.path.exists(
                    get_file_path(ddp_tmpdir, idx=i, epoch=epoch,
                                  is_train=False)), f"val sample {i} was not accessed during epoch {epoch}"
            # the eval dataloader is spun once more to initialize the rng, so expecting num_epochs + 1 to not exist
            assert not os.path.exists(get_file_path(ddp_tmpdir, idx=i, epoch=num_epochs + 1,
                                                    is_train=False)), f"val sample {i} was accessed too many times"

        is_train_to_pickles: Dict[bool, List[Dict[str, types.Tensor]]] = {True: [], False: []}

        for epoch in range(num_epochs):
            for local_rank in range(trainer.device.nproc_per_node):
                for is_train in (True, False):
                    data: Dict[str, types.Tensor] = torch.load(  # type: ignore
                        get_batch_file_path(ddp_tmpdir, rank=local_rank, epoch=epoch, is_train=is_train),
                        map_location='cpu',
                    )
                    for pickle in is_train_to_pickles[is_train]:
                        assert not torch.all(data['last_input'] == pickle['last_input'])
                        assert not torch.all(data['last_target'] == pickle['last_target'])
                    is_train_to_pickles[is_train].append(data)


def test_ddp_cuda_available_check(mosaic_trainer_hparams: TrainerHparams):
    with mock.patch.object(torch.cuda, 'device_count') as device_count, \
        mock.patch.object(torch.cuda, 'is_available') as is_cuda_available:
        is_cuda_available.return_value = False
        device_count = 1

        mosaic_trainer_hparams.device = GPUDeviceHparams(n_gpus=1)
        assert (not torch.cuda.is_available())

        with pytest.raises(ValueError, match="CUDA not available but gpu backend requested."):
            mosaic_trainer_hparams.initialize_object()


def test_ddp_cuda_ngpus_check(mosaic_trainer_hparams: TrainerHparams):
    with mock.patch.object(torch.cuda, 'device_count') as device_count, \
        mock.patch.object(torch.cuda, 'is_available') as is_cuda_available:
        is_cuda_available.return_value = True
        device_count.return_value = 2

        mosaic_trainer_hparams.device = GPUDeviceHparams(n_gpus=8)

        with pytest.raises(ValueError, match="Requested 8 GPUs, but only 2 available."):
            mosaic_trainer_hparams.initialize_object()


def test_ddp_nccl_check(mosaic_trainer_hparams: TrainerHparams):
    with mock.patch.object(torch.cuda, 'device_count') as device_count, \
        mock.patch.object(torch.distributed, 'is_nccl_available') as nccl_available, \
        mock.patch.object(torch.cuda, 'is_available') as is_cuda_available:

        device_count.return_value = 1
        is_cuda_available.return_value = True
        nccl_available.return_value = False

        mosaic_trainer_hparams.device = GPUDeviceHparams(n_gpus=1)

        with pytest.raises(ValueError, match="Requested NCCL backend not available in torch.distributed"):
            mosaic_trainer_hparams.initialize_object()



# Copyright 2021 MosaicML. All Rights Reserved.

from unittest.mock import MagicMock

import pytest

from composer.callbacks import GradMonitorHparams
from composer.datasets.synthetic import SyntheticDatasetHparams
from composer.trainer import TrainerHparams


def _do_trainer_fit(mosaic_trainer_hparams: TrainerHparams, log_layers=False):
    grad_monitor_hparams = GradMonitorHparams(log_layer_grad_norms=log_layers)
    mosaic_trainer_hparams.callbacks.append(grad_monitor_hparams)
    mosaic_trainer_hparams.ddp.fork_rank_0 = False
    mosaic_trainer_hparams.max_epochs = 1

    mosaic_trainer_hparams.total_batch_size = 50
    trainer = mosaic_trainer_hparams.initialize_object()
    log_destination = MagicMock()
    log_destination.will_log.return_value = True
    trainer.logger.backends = [log_destination]
    trainer.fit()

    assert isinstance(mosaic_trainer_hparams.train_dataset, SyntheticDatasetHparams)
    num_train_samples = mosaic_trainer_hparams.train_dataset.sample_pool_size
    num_train_steps = num_train_samples // mosaic_trainer_hparams.total_batch_size

    return log_destination, num_train_steps


@pytest.mark.timeout(60)
@pytest.mark.run_long
def test_grad_monitor_no_layers(mosaic_trainer_hparams: TrainerHparams):
    log_destination, num_train_steps = _do_trainer_fit(mosaic_trainer_hparams, log_layers=False)
    grad_norm_calls = 0
    for log_call in log_destination.log_metric.mock_calls:
        metrics = log_call[1][3]
        if "grad_l2_norm/step" in metrics:
            grad_norm_calls += 1

    # expected to log gradient norm once per step (total batch)
    assert grad_norm_calls == num_train_steps


@pytest.mark.timeout(60)
@pytest.mark.run_long
def test_grad_moniter_per_layer(mosaic_trainer_hparams: TrainerHparams):
    log_destination, num_train_steps = _do_trainer_fit(mosaic_trainer_hparams, log_layers=True)
    layer_norm_calls = 0
    for log_call in log_destination.log_metric.mock_calls:
        metrics = log_call[1][3]
        if not isinstance(metrics, dict):
            continue
        if any(map(lambda x: "layer_grad_l2_norm" in x, metrics.keys())):
            layer_norm_calls += 1

    # expected to log layer grad norms once per training step
    assert layer_norm_calls == num_train_steps



# Copyright 2021 MosaicML. All Rights Reserved.

import pytest
import torch
import torch.nn.functional as F

from composer.algorithms import StochasticDepth, StochasticDepthHparams
from composer.algorithms.stochastic_depth.sample_stochastic_layers import SampleStochasticBottleneck
from composer.algorithms.stochastic_depth.stochastic_depth import STOCHASTIC_LAYER_MAPPING
from composer.algorithms.stochastic_depth.stochastic_layers import StochasticBottleneck, _sample_bernoulli
from composer.core import Event, Logger, State, surgery
from composer.core.types import Precision
from composer.datasets import SyntheticDatasetHparams
from composer.loggers import Logger
from composer.models import ResNet50Hparams
from tests.fixtures.dummy_fixtures import get_dataloader


@pytest.fixture()
def dummy_state(dummy_dataloader_hparams):
    model = ResNet50Hparams(num_classes=100).initialize_object()
    dataset_hparams = SyntheticDatasetHparams(num_classes=10,
                                              shape=[3, 32, 32],
                                              sample_pool_size=1000000,
                                              device="cpu",
                                              one_hot=False,
                                              drop_last=True,
                                              shuffle=False)
    train_dataloader = get_dataloader(dataset_hparams.initialize_object(), dummy_dataloader_hparams)
    return State(epoch=50,
                 step=50,
                 train_dataloader=train_dataloader,
                 train_batch_size=100,
                 eval_batch_size=100,
                 grad_accum=1,
                 max_epochs=100,
                 model=model,
                 precision=Precision.FP32)


@pytest.mark.parametrize('stochastic_method', ['block', 'sample'])
@pytest.mark.parametrize('target_layer_name', ['ResNetBottleneck'])
def test_stochastic_depth_bottleneck_replacement(dummy_state, stochastic_method, target_layer_name, noop_dummy_logger):
    target_layer, stochastic_layer = STOCHASTIC_LAYER_MAPPING[stochastic_method][target_layer_name]
    target_block_count = surgery.count_module_instances(dummy_state.model, target_layer)

    sd = StochasticDepth(stochastic_method=stochastic_method,
                         target_layer_name=target_layer_name,
                         drop_rate=0.5,
                         drop_distribution='linear',
                         drop_warmup=0.0,
                         use_same_gpu_seed=False)
    sd.apply(Event.INIT, dummy_state, noop_dummy_logger)
    stochastic_block_count = surgery.count_module_instances(dummy_state.model, stochastic_layer)

    assert target_block_count == stochastic_block_count


@pytest.fixture
def device_ids():
    return torch.arange(8)


@pytest.fixture
def module_ids():
    return torch.arange(16)


@pytest.fixture
def probability():
    return 0.5


def test_sample_bernoulli_use_same_gpu_seed(probability, device_ids, module_ids):
    mask_matrix = torch.zeros(len(device_ids), len(module_ids))
    for device_id in device_ids:
        torch.manual_seed(0)  # Simulates each device having the same random_seed as is the case with DDP
        for module_id in module_ids:
            generator = torch.Generator().manual_seed(144385)
            mask_matrix[device_id, module_id] = _sample_bernoulli(probability=torch.tensor(probability),
                                                                  device_id=device_id.item(),
                                                                  module_id=module_id.item(),
                                                                  num_modules=len(module_ids),
                                                                  generator=generator,
                                                                  use_same_gpu_seed=True,
                                                                  use_same_depth_across_gpus=True)

    assert torch.unique(mask_matrix, dim=0).size(0) == 1


def test_sample_bernoulli_devices_not_same_gpu(probability, device_ids, module_ids):
    mask_matrix = torch.zeros(len(device_ids), len(module_ids))
    for device_id in device_ids:
        torch.manual_seed(0)  # Simulates each device having the same random_seed as is the case with DDP
        for module_id in module_ids:
            generator = torch.Generator().manual_seed(144385)
            mask_matrix[device_id, module_id] = _sample_bernoulli(probability=torch.tensor(probability),
                                                                  device_id=device_id.item(),
                                                                  module_id=module_id.item(),
                                                                  num_modules=len(module_ids),
                                                                  generator=generator,
                                                                  use_same_gpu_seed=False,
                                                                  use_same_depth_across_gpus=False)
    # Check for unique drop masks across devices
    assert (torch.unique(mask_matrix, dim=0).size(0) != 1)


def test_sample_bernoulli_layers_not_same_gpu(probability, device_ids, module_ids):
    mask_matrix = torch.zeros(len(device_ids), len(module_ids))
    for device_id in device_ids:
        torch.manual_seed(0)  # Simulates each device having the same random_seed as is the case with DDP
        for module_id in module_ids:
            generator = torch.Generator().manual_seed(144385)
            mask_matrix[device_id, module_id] = _sample_bernoulli(probability=torch.tensor(probability),
                                                                  device_id=device_id.item(),
                                                                  module_id=module_id.item(),
                                                                  num_modules=len(module_ids),
                                                                  generator=generator,
                                                                  use_same_gpu_seed=False,
                                                                  use_same_depth_across_gpus=False)
    # Check for unique drop masks across layers
    assert (torch.unique(mask_matrix, dim=1).size(1) != 1)


@pytest.mark.parametrize('drop_rate', [1.0])
def test_stochastic_bottleneck_drop(drop_rate):
    X = torch.randn(4, 4, 16, 16)
    generator = torch.Generator().manual_seed(144385)
    stochastic_layer = StochasticBottleneck(drop_rate=drop_rate,
                                            module_id=2,
                                            module_count=10,
                                            use_same_depth_across_gpus=False,
                                            use_same_gpu_seed=False,
                                            rand_generator=generator,
                                            inplanes=4,
                                            planes=1)
    new_X = stochastic_layer(X)
    assert new_X is X


@pytest.mark.parametrize('drop_rate', [0.0])
def test_stochastic_bottleneck_keep(drop_rate):
    X = torch.randn(4, 4, 16, 16)
    generator = torch.Generator().manual_seed(144385)
    stochastic_layer = StochasticBottleneck(drop_rate=drop_rate,
                                            module_id=2,
                                            module_count=10,
                                            use_same_depth_across_gpus=False,
                                            use_same_gpu_seed=False,
                                            rand_generator=generator,
                                            inplanes=4,
                                            planes=1)
    new_X = stochastic_layer(X)
    assert new_X is not X


@pytest.mark.parametrize('drop_rate', [1.0])
def test_sample_stochastic_bottleneck_drop_all(drop_rate):
    x = F.relu(torch.randn(4, 4, 16, 16))  # inputs and outputs will match if the input has been ReLUed
    sample_stochastic_block = SampleStochasticBottleneck(drop_rate=drop_rate, inplanes=4, planes=1)
    x_dropped = sample_stochastic_block(x)
    assert torch.all(x == x_dropped)


@pytest.fixture()
def target_layer_name() -> str:
    return "ResNetBottleneck"


@pytest.fixture()
def stochastic_method():
    return "sample"


@pytest.fixture
def dummy_hparams(stochastic_method, target_layer_name, drop_rate, drop_distribution, drop_warmup, use_same_gpu_seed):
    return StochasticDepthHparams(stochastic_method, target_layer_name, drop_rate, drop_distribution, drop_warmup,
                                  use_same_gpu_seed)


def get_drop_rate_list(module: torch.nn.Module, drop_rates=[]):
    if (len(list(module.children())) == 0 and len(list(module.parameters())) > 0):
        return
    else:
        for name, child in module.named_children():
            if hasattr(child, 'drop_rate'):
                drop_rates.append(child.drop_rate)
            get_drop_rate_list(child, drop_rates)


@pytest.mark.parametrize("step", [50, 100, 1000])
@pytest.mark.parametrize("drop_rate", [0.0, 0.5, 1.0])
@pytest.mark.parametrize("drop_distribution", ['uniform', 'linear'])
@pytest.mark.parametrize("use_same_gpu_seed", [True])
@pytest.mark.parametrize("drop_warmup", [0.1])
def test_drop_rate_warmup(step, dummy_hparams, dummy_state, noop_dummy_logger: Logger):
    dummy_algorithm = dummy_hparams.initialize_object()
    old_drop_rates = []
    get_drop_rate_list(dummy_state.model, drop_rates=old_drop_rates)
    dummy_state.step = step
    dummy_algorithm.apply(Event.BATCH_START, dummy_state, noop_dummy_logger)
    new_drop_rates = []
    get_drop_rate_list(dummy_state.model, drop_rates=new_drop_rates)

    drop_warmup_iters = dummy_state.steps_per_epoch * dummy_state.max_epochs * dummy_algorithm.hparams.drop_warmup
    assert torch.all(torch.tensor(new_drop_rates) == ((step / drop_warmup_iters) * torch.tensor(old_drop_rates)))


@pytest.mark.parametrize("stochastic_method", ['nonsense'])
def test_invalid_method_name(stochastic_method, target_layer_name):
    with pytest.raises(ValueError):
        sd_hparams = StochasticDepthHparams(stochastic_method=stochastic_method, target_layer_name=target_layer_name)
        sd_hparams.validate()


@pytest.mark.parametrize("target_layer_name", ['nonsense_pt2'])
def test_invalid_layer_name(stochastic_method, target_layer_name):
    with pytest.raises(ValueError):
        sd_hparams = StochasticDepthHparams(stochastic_method=stochastic_method, target_layer_name=target_layer_name)
        sd_hparams.validate()


@pytest.mark.parametrize("drop_rate", [-0.5, 1.7])
def test_invalid_drop_rate(stochastic_method, target_layer_name, drop_rate):
    with pytest.raises(ValueError):
        sd_hparams = StochasticDepthHparams(stochastic_method=stochastic_method,
                                            target_layer_name=target_layer_name,
                                            drop_rate=drop_rate)
        sd_hparams.validate()


@pytest.mark.parametrize("drop_distribution", ['nonsense_pt3'])
def test_invalid_drop_distribution(stochastic_method, target_layer_name, drop_distribution):
    with pytest.raises(ValueError):
        sd_hparams = StochasticDepthHparams(stochastic_method=stochastic_method,
                                            target_layer_name=target_layer_name,
                                            drop_distribution=drop_distribution)
        sd_hparams.validate()


@pytest.mark.parametrize("drop_warmup", [-0.5, 1.7])
def test_invalid_drop_warmup(stochastic_method, target_layer_name, drop_warmup):
    with pytest.raises(ValueError):
        sd_hparams = StochasticDepthHparams(stochastic_method,
                                            target_layer_name=target_layer_name,
                                            drop_warmup=drop_warmup)
        sd_hparams.validate()



# Copyright 2021 MosaicML. All Rights Reserved.

"""
Functional API

Functional forms of methods are available here via::

    from composer import functional as CF
    my_model = CF.apply_blurpool(my_model)


"""
from composer.algorithms.augmix import augment_and_mix as augment_and_mix
from composer.algorithms.blurpool import apply_blurpool as apply_blurpool
from composer.algorithms.colout.colout import colout as colout
from composer.algorithms.curriculum_learning.curriculum_learning import apply_curriculum as apply_curriculum
from composer.algorithms.cutout.cutout import cutout as cutout
from composer.algorithms.label_smoothing import smooth_labels as smooth_labels
from composer.algorithms.layer_freezing import freeze_layers as freeze_layers
from composer.algorithms.mixup import mixup_batch as mixup_batch
from composer.algorithms.progressive_resizing import resize_inputs as resize_inputs
from composer.algorithms.randaugment import randaugment as randaugment
from composer.algorithms.scale_schedule.scale_schedule import scale_scheduler as scale_scheduler
from composer.algorithms.selective_backprop.selective_backprop import do_selective_backprop as do_selective_backprop
from composer.algorithms.selective_backprop.selective_backprop import selective_backprop as selective_backprop
from composer.algorithms.squeeze_excite import apply_se as apply_se



# Copyright 2021 MosaicML. All Rights Reserved.

"""Test Ghost Batch Normalization, both as an algorithm and module
"""

import math
from typing import cast
from unittest.mock import MagicMock, Mock

import pytest
import torch

from composer.algorithms import ghost_batchnorm as ghostbn
from composer.algorithms.ghost_batchnorm.ghost_batchnorm import GhostBatchNormHparams, _GhostBatchNorm
from composer.core import Event, State, surgery
from composer.trainer import TrainerHparams
from tests.fixtures.dummy_fixtures import logger_mock as logger_mock
from tests.utils.trainer_fit import train_model

_GHOSTBN_MODULE_CLASS = _GhostBatchNorm
_GHOSTBN_CORRECT_EVENT = Event.INIT

_TEST_NUM_DIMS = [1, 2, 3]
_TEST_GHOST_BATCH_SIZES = [1, 2, 3, 5]
_TEST_BATCH_SIZES = [12]  # multiple of some, but not all, ghost batch sizes


class ModuleWithBatchnorm(torch.nn.Module):

    def __init__(self, num_dims, num_features=4):
        super().__init__()
        eps = 0  # makes it easier to check normalization correctness
        factory_func = {
            1: torch.nn.BatchNorm1d,
            2: torch.nn.BatchNorm2d,
            3: torch.nn.BatchNorm3d,
        }
        self.bn = factory_func[num_dims](num_features, eps=eps)
        self.num_dims = num_dims
        self.num_features = num_features
        self.non_batchnorm_module = torch.nn.Conv2d(4, 5, (1, 1))

    def forward(self, input: torch.Tensor):
        return self.bn(input)


@pytest.fixture
def state(num_dims: int) -> State:
    return MagicMock(model=ModuleWithBatchnorm(num_dims=num_dims))


@pytest.fixture
def algo_instance(ghost_batch_size: int):
    hparams = ghostbn.GhostBatchNormHparams(ghost_batch_size=ghost_batch_size)
    return hparams.initialize_object()


@pytest.mark.parametrize('num_dims', [
    1, 2, 3,
    pytest.param(4, marks=pytest.mark.xfail(raises=KeyError)),
    pytest.param(-1, marks=pytest.mark.xfail(raises=KeyError))
])
def test_batchnorm_gets_replaced_functional(num_dims: int):
    """GhostBatchNorm{1,2,3}d should work, but other ints should throw"""
    module = ModuleWithBatchnorm(num_dims)
    assert surgery.count_module_instances(module, _GHOSTBN_MODULE_CLASS) == 0
    ghostbn.apply_ghost_batchnorm(module, ghost_batch_size=1)
    assert surgery.count_module_instances(module, _GHOSTBN_MODULE_CLASS) == 1


@pytest.mark.parametrize('num_dims', _TEST_NUM_DIMS)
@pytest.mark.parametrize('ghost_batch_size', _TEST_GHOST_BATCH_SIZES)
@pytest.mark.parametrize('batch_size', _TEST_BATCH_SIZES)
class TestGhostBatchesNormalized:

    def _assert_ghost_batches_normalized(self, module: ModuleWithBatchnorm, ghost_batch_size: int,
                                         batch_size: int) -> None:
        torch.manual_seed(123)
        size = [batch_size, module.num_features] + ([3] * module.num_dims)
        X = torch.randn(size=size)
        module.train()
        out = module(X)
        n_channels = out.shape[1]
        # reduce over everything but channel idx
        reduce_dims = (0,) + tuple(range(2, out.ndim))

        nchunks = int(math.ceil(batch_size / ghost_batch_size))
        for ghost_batch in out.chunk(nchunks):
            channel_variances, channel_means = torch.var_mean(ghost_batch, dim=reduce_dims, unbiased=False)
            torch.testing.assert_close(channel_variances, torch.ones(n_channels))
            torch.testing.assert_close(channel_means, torch.zeros(n_channels))

    def test_normalization_correct_functional(self, num_dims: int, ghost_batch_size: int, batch_size: int) -> None:
        module = ModuleWithBatchnorm(num_dims=num_dims)
        ghostbn.apply_ghost_batchnorm(module, ghost_batch_size=ghost_batch_size)
        self._assert_ghost_batches_normalized(module=module, ghost_batch_size=ghost_batch_size, batch_size=batch_size)

    def test_normalization_correct_algorithm(self, state, algo_instance, num_dims: int, ghost_batch_size: int,
                                             batch_size: int, logger_mock) -> None:
        algo_instance.apply(_GHOSTBN_CORRECT_EVENT, state, logger_mock)
        module = cast(ModuleWithBatchnorm, state.model)
        self._assert_ghost_batches_normalized(module=module, ghost_batch_size=ghost_batch_size, batch_size=batch_size)


@pytest.mark.parametrize('ghost_batch_size', [4])
def test_correct_event_matches(algo_instance):
    assert algo_instance.match(_GHOSTBN_CORRECT_EVENT, Mock(side_effect=ValueError))


@pytest.mark.parametrize('ghost_batch_size', [4])
@pytest.mark.parametrize('event', Event)  # enum iteration
def test_incorrect_event_does_not_match(event: Event, algo_instance):
    if event == _GHOSTBN_CORRECT_EVENT:
        return
    assert not algo_instance.match(event, Mock(side_effect=ValueError))


@pytest.mark.parametrize('ghost_batch_size', [4])
@pytest.mark.parametrize('num_dims', [2])
def test_algorithm_logging(logger_mock, state, algo_instance):
    algo_instance.apply(Event.INIT, state, logger_mock)
    logger_mock.metric_fit.assert_called_once_with({
        'GhostBatchNorm/num_new_modules': 1,
    })


@pytest.mark.run_long
@pytest.mark.timeout(90)
def test_ghost_batchnorm_trains(mosaic_trainer_hparams: TrainerHparams):
    mosaic_trainer_hparams.algorithms = [GhostBatchNormHparams(ghost_batch_size=16)]
    train_model(mosaic_trainer_hparams)



# Copyright 2021 MosaicML. All Rights Reserved.

import numpy as np
import pytest
import torch

from composer.algorithms import ChannelsLastHparams
from composer.core.event import Event
from composer.core.state import State
from composer.core.types import Model, Precision, Tensor


def _has_singleton_dimension(tensor: Tensor) -> bool:
    return any(s == 1 for s in tensor.shape)


def _infer_memory_format(tensor: Tensor) -> str:
    if _has_singleton_dimension(tensor):
        raise ValueError(f'Tensor of shape {tensor.shape} has singleton dimensions, '
                         'memory format cannot be infered from strides.')
    base_order = list('nchw')  # type: ignore

    strides = tensor.stride()
    if isinstance(strides, tuple) and len(strides) == 4:
        order = np.argsort(strides)
        # smallest stride should be last in format, so reverse order
        memory_format = ''.join([base_order[o] for o in reversed(order)])
        return memory_format
    else:
        raise ValueError(f'Tensor must be 4-dim, got shape {tensor.shape}')


@pytest.fixture()
def state(simple_conv_model: Model):
    return State(
        model=simple_conv_model,
        train_batch_size=100,
        eval_batch_size=100,
        precision=Precision.FP32,
        grad_accum=1,
        max_epochs=10,
    )


def test_channels_last_algorithm(state, dummy_logger):
    channels_last = ChannelsLastHparams().initialize_object()

    assert state.model is not None
    assert _infer_memory_format(state.model.module.conv1.weight) == 'nchw'
    channels_last.apply(Event.TRAINING_START, state, dummy_logger)
    assert _infer_memory_format(state.model.module.conv1.weight) == 'nhwc'


"""
Test helper utility _infer_memory_format
"""


@pytest.fixture(params=[True, False])
def tensor(request) -> Tensor:
    strided = request.param
    tensor = torch.randn((16, 32, 32, 64))
    if strided:
        tensor = tensor[::2, ::2, ::2, ::2]
    return tensor


def test_infer_memory_format_nhwc(tensor):
    tensor = tensor.to(memory_format=torch.channels_last)
    assert _infer_memory_format(tensor) == 'nhwc'


def test_infer_memory_format_nchw(tensor):
    tensor = tensor.to(memory_format=torch.contiguous_format)
    assert _infer_memory_format(tensor) == 'nchw'


def test_infer_memory_format_wcnh(tensor):
    tensor = tensor.to(memory_format=torch.contiguous_format)
    tensor = tensor.permute(2, 1, 3, 0)
    assert _infer_memory_format(tensor) == 'wcnh'


def test_infer_memory_format_incorrect_ndims():
    tensor = torch.randn((16, 32, 32))
    with pytest.raises(ValueError):
        _infer_memory_format(tensor)


def test_infer_memory_format_singleton():
    tensor = torch.randn((16, 32, 1, 64))
    with pytest.raises(ValueError):
        _infer_memory_format(tensor)



# Copyright 2021 MosaicML. All Rights Reserved.

import torch
import torch.nn as nn


# TODO: refactor as layer_types and add tests
def kaiming_normal(w: nn.Module):
    if isinstance(w, torch.nn.Linear) or isinstance(w, torch.nn.Conv2d):
        torch.nn.init.kaiming_normal_(w.weight)


def kaiming_uniform(w: nn.Module):
    if isinstance(w, torch.nn.Linear) or isinstance(w, torch.nn.Conv2d):
        torch.nn.init.kaiming_uniform_(w.weight)


def xavier_uniform(w: nn.Module):
    if isinstance(w, torch.nn.Linear) or isinstance(w, torch.nn.Conv2d):
        torch.nn.init.xavier_uniform_(w.weight)


def bn_ones(w: nn.Module):
    if isinstance(w, torch.nn.BatchNorm2d):
        w.weight.data = torch.ones_like(w.weight.data.shape)
        w.bias.data = torch.zeros_like(w.bias.data)


def bn_uniform(w: nn.Module):
    if isinstance(w, torch.nn.BatchNorm2d):
        w.weight.data = torch.rand(w.weight.data.shape)
        w.bias.data = torch.zeros_like(w.bias.data)


def get_initializer(name: str):
    initializer_dict = {
        "kaiming_normal": kaiming_normal,
        "kaiming_uniform": kaiming_uniform,
        "bn_uniform": bn_uniform,
        "bn_ones": bn_ones,
        "xavier_uniform": xavier_uniform
    }
    if name not in initializer_dict:
        raise ValueError(f"Initializer '{name}' not found.")
    return initializer_dict[name]



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

import yahp as hp

import composer
from composer.core.algorithm import Algorithm


@dataclass
class AlgorithmHparams(hp.Hparams, ABC):

    @abstractmethod
    def initialize_object(self) -> Algorithm:
        pass

    def __init_subclass__(cls) -> None:
        return super().__init_subclass__()

    @classmethod
    def load(cls, alg_params: Optional[str] = None) -> AlgorithmHparams:
        from composer.algorithms.algorithm_registry import get_algorithm_registry
        registry = get_algorithm_registry()
        inverted_registry = {v: k for (k, v) in registry.items()}
        alg_name = inverted_registry[cls]
        alg_folder = os.path.join(os.path.dirname(composer.__file__), "yamls", "algorithms")
        if alg_params is None:
            hparams_file = os.path.join(alg_folder, f"{alg_name}.yaml")
        else:
            hparams_file = os.path.join(alg_folder, alg_name, f"{alg_params}.yaml")
        if os.path.exists(hparams_file):
            alg_hparams = cls.create(hparams_file)
            assert isinstance(alg_hparams, AlgorithmHparams), "hparams.create should return an instance of its type"
            return alg_hparams
        return cls()

    @classmethod
    def load_multiple(cls, *algorithms: str):
        from composer.algorithms.algorithm_registry import get_algorithm_registry
        registry = get_algorithm_registry()
        alg_hparams = []
        for alg in algorithms:
            alg_parts = alg.split("/")
            alg_name = alg_parts[0]
            if len(alg_parts) > 1:
                alg_params = "/".join(alg_parts[1:])
            else:
                alg_params = None
            try:
                alg_param = registry[alg_name]
            except KeyError as e:
                raise ValueError(f"Algorithm {e.args[0]} not found") from e
            alg_hparams.append(alg_param.load(alg_params))
        return alg_hparams



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Mapping, Tuple

import transformers

from composer.models.base import BaseMosaicModel
from composer.models.nlp_metrics import LanguageCrossEntropyLoss

if TYPE_CHECKING:
    from composer.core.types import Batch, Metrics, Tensors

log = logging.getLogger(__name__)


class MosaicTransformer(BaseMosaicModel):
    """
    Implements the base logic that all Transformers can build on top off.

    Args:
        module (transformers.PreTrainedModel): An instance of PreTrainedModel that contains the forward pass function.
        config (transformers.PretrainedConfig): The PretrainedConfig object that stores information about the model
                                                hyperparameters.
        tokenizer_name (str): The name of the tokenizer used for tihs model, necessary to assert required model inputs.
    """

    def __init__(self, module: transformers.PreTrainedModel, config: transformers.PretrainedConfig,
                 tokenizer_name: str) -> None:
        super().__init__()
        self.module = module
        self.config = config
        self.tokenizer = transformers.AutoTokenizer.from_pretrained(tokenizer_name)
        log.info("Number of parameters in the model: " \
                 f"{sum(p.numel() for p in module.parameters()):,}")  # type: ignore (thirdparty)
        log.info("Number of trainable parameters in the model: "
                 f"{sum(p.numel() for p in module.parameters() if p.requires_grad):,}")  # type: ignore (thirdparty)

        # the set of inputs that a model expects
        # if an algorithm modifies the loss function, it must remove "labels" from this set.
        self.model_inputs = set(self.tokenizer.model_input_names)
        self.model_inputs.update(set({"labels"}))

        # define metrics for measurements
        self.train_loss = LanguageCrossEntropyLoss()
        self.val_loss = LanguageCrossEntropyLoss()

    def loss(self, outputs: Mapping, batch: Batch) -> Tensors:
        """
        Computes the loss of the tensor from the output.

        Args:
            outputs (Mapping): The dictionary output from the model. It could contain the loss as computed by HF, or
                               algorithms can pop the labels from the input in case they modify the loss function.
            batch (Batch): the set of ground truth labels to use to compute the loss against.

        Returns:
            None

        Raises:
            NotImplementedError: a model-specific and task-specific loss function must be written.

        Notes:
            We don't implement this for the generic Transformer abstraction, since loss functions are model and
            objective specific. A single model architeture could use a myriad of loss functions which are better left
            epressed by the user.
        """

        raise NotImplementedError("A model-specific loss function must be written.")

    def forward(self, batch: Batch) -> Mapping:
        """
        Runs the forward pass of the model.

        Args:
            batch (Batch): a dictionary of Dict[str, Tensor] of inputs that the model expects, as found in
            MosaicTransformer.get_model_inputs().

        Returns:
            output (Mapping): a dictionary of model outputs. It will include the loss if `labels` is passed as an input.
        """
        if not isinstance(batch, dict):
            raise ValueError(f'Model expects batch to be a dict, got {type(batch)}')

        for key in self.model_inputs:
            if key not in batch.keys():
                raise ValueError(f'Batch missing key: {key}')

        output = self.module(**batch)  # type: ignore (thirdparty)
        return output

    def metrics(self, train: bool = False) -> Metrics:
        """
        Returns the Metrics objects for computing the training and validation losses.
        Downstream models should override this method if they would like to add task-specific metrics.


        Args:
            train (bool): a boolean flag to indicate whether to return training or validation metrics.

        Returns:
            A  Metrics object that can be used to calculate task performance.

        Notes:
            If train=True, then it might calculate the training loss twice if algorithms are overriding the loss fn.
            This could be expensive due to the computational cost of softmax; it is worth exploring caching stratgies.
        """
        return self.train_loss if train else self.val_loss

    def validate(self, batch: Batch) -> Tuple[Mapping, None]:
        """
        Runs the validation step.

        Args:
            batch (Batch): a dictionary of Dict[str, Tensor] of inputs that the model expects, as found in
            MosaicTransformer.get_model_inputs().

        Returns:
            A tuple of (Mapping, None) with the output from the forward pass. This is fed into a Metrics object.
        """

        assert self.training is False, "For validation, model must be in eval mode"
        output = self.forward(batch)

        return (output, None)

    def get_model_inputs(self):
        """
        Returns a set of inputs that the model expects in the forward pass.

        Args:
            None

        Returns:
            self.modellinputs (set): the set of keys that are expected in the Mapping used to compute the forward pass.

        Notes:
            If an algorithm wants to interact with the model inputs (for instance, popping the labels for a custom loss
            fn, or adding attention head masks for head pruning, it must access self.set_model_inputs().
        """

        return self.model_inputs



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.trainer import devices as devices
from composer.trainer.ddp import DDPDataLoader as DDPDataLoader
from composer.trainer.ddp import DDPHparams as DDPHparams
from composer.trainer.ddp import FileStoreHparams as FileStoreHparams
from composer.trainer.ddp import StoreHparams as StoreHparams
from composer.trainer.ddp import TCPStoreHparams as TCPStoreHparams
from composer.trainer.trainer import Trainer as Trainer
from composer.trainer.trainer_hparams import TrainerHparams as TrainerHparams

load = TrainerHparams.load



# Copyright 2021 MosaicML. All Rights Reserved.

import pytest
import torch

from composer.algorithms import CutOutHparams
from composer.algorithms.cutout.cutout import generate_mask
from composer.core.types import Event, Tensor
from composer.trainer.trainer_hparams import TrainerHparams
from tests.utils.trainer_fit import train_model


def _is_square(cutout_box: Tensor) -> bool:
    height, width = cutout_box.size()

    return height == width


# Box validaton checks for a continuous rectangle, cannot handle multiple/coalesced boxes along x, y dimensions
def _box_validate(mask_box: Tensor) -> None:
    # Box is not contiguous if there are any 0's in the tensor
    box_is_contiguous = not (0 in mask_box)
    assert box_is_contiguous


def _find_box(img_2d: Tensor) -> Tensor:
    height, width = img_2d.size()

    # Generate helper tensors
    ones = torch.ones(height, width)
    zeros = torch.zeros(height, width)

    # Find the box
    # First create h x w filter populated with ones where it thinks there's a box, then find coordinates of ones
    filter_box = torch.where(img_2d == 0, ones, zeros)
    box_x, box_y = torch.nonzero(filter_box, as_tuple=True)  # Find all points where filter_box is 1

    # Check for no box found
    if ((box_x.size()[0], box_y.size()[0]) == (0, 0)):
        # Return valid box as this is possible when cutout_length=1
        return torch.ones(1, 1)
    else:
        # Returns box defined by longest diagonal
        return filter_box[box_x[0]:box_x[-1] + 1, box_y[0]:box_y[-1] + 1]


def check_box(batch_size, channels, input):
    for b in range(batch_size):
        for c in range(channels):
            mask_box = _find_box(input[b, c, :, :])
            _box_validate(mask_box)


# Test square, rectangle inputs
@pytest.fixture(params=[(1, 1, 16, 16), (1, 1, 16, 32)])
def tensor_sizes(request):
    return request.param


# cutout_length=1 won't 0 out (no box is valid)
# cutout_length=3 should produce 2x2 box due to floor division except when boundary clipping
# cutout_length=4 should produce 4x4 box due except when boundary clipping
@pytest.fixture(params=[1, 3, 4])
def cutout_length(request):
    return request.param


# Check corners, edges and middle
@pytest.fixture(params=[(0, 0), (16, 0), (0, 16), (16, 16), (7, 7)])
def anchors(request):
    return request.param


def test_cutout_mask(tensor_sizes, cutout_length, anchors):

    batch_size, channels, height, width = tensor_sizes
    x, y = anchors

    test_mask = torch.ones(tensor_sizes)
    test_mask = generate_mask(mask=test_mask, width=width, height=height, x=x, y=y, cutout_length=cutout_length)

    check_box(batch_size, channels, test_mask)


@pytest.mark.parametrize('batch_size', [1, 4])
@pytest.mark.parametrize('channels', [1, 4])
@pytest.mark.parametrize('height', [32, 64])
@pytest.mark.parametrize('width', [32, 71])
@pytest.mark.parametrize('cutout_length', [16, 23])
def test_cutout_algorithm(batch_size, channels, height, width, cutout_length, dummy_logger, dummy_state):

    # Initialize input tensor
    #   - Add bias to prevent 0. pixels, causes check_box() to fail since based on search for 0's
    #   - Real data can have 0. pixels but this will not affect cutout algorithm since mask is generated independent of input data
    input = torch.rand((batch_size, channels, height, width)) + 1

    # Fix cutout_n_holes=1, mask generation is additive and box validation isn't smart enough to detect multiple/coalesced boxes
    algorithm = CutOutHparams(n_holes=1, length=cutout_length).initialize_object()
    state = dummy_state
    state.batch = (input, torch.Tensor())

    algorithm.apply(Event.AFTER_DATALOADER, state, dummy_logger)

    input, _ = state.batch
    check_box(batch_size, channels, input)


@pytest.mark.run_long
@pytest.mark.timeout(90)
def test_cutout_trains(mosaic_trainer_hparams: TrainerHparams):
    mosaic_trainer_hparams.algorithms = [CutOutHparams(n_holes=1, length=4)]
    train_model(mosaic_trainer_hparams)



# Copyright 2021 MosaicML. All Rights Reserved.

import contextlib
from typing import Callable, ContextManager, Union

import torch
import torch.cuda.amp

from composer.core.precision import Precision


def default_precision_factory() -> Callable[[Union[str, Precision]], ContextManager]:
    """Returns a context manager to automatically cast to a specific precision

    Args:
        precision (str or Precision): Precision for the context
    """
    if torch.cuda.is_available():
        return lambda precision: torch.cuda.amp.autocast(Precision(precision) == Precision.AMP)
    else:

        def null(precision):
            assert Precision(
                precision) != Precision.AMP, "Precision AMP is only available when `torch.cuda.is_available() == True`."
            return contextlib.nullcontext()

        return null


# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.alibi import gpt2_alibi as gpt2_alibi
from composer.algorithms.alibi.alibi import Alibi as Alibi
from composer.algorithms.alibi.alibi import AlibiHparams as AlibiHparams
from composer.algorithms.alibi.alibi import apply_alibi as apply_alibi



# Copyright 2021 MosaicML. All Rights Reserved.

import pytest
import torch

from composer.algorithms import SqueezeExcite, SqueezeExciteConv2d
from composer.algorithms.squeeze_excite.squeeze_excite import SqueezeExciteHparams
from composer.core import Event, Logger, State, surgery
from composer.core.types import Tensors
from composer.trainer.trainer_hparams import TrainerHparams
from tests.utils.trainer_fit import train_model


def _do_squeeze_excite(state_with_model: State, simple_conv_model_input: Tensors, logger: Logger):
    batch = (simple_conv_model_input, None)
    out = state_with_model.model.forward(batch)
    original_size = out.size()
    original_conv_count = surgery.count_module_instances(state_with_model.model, torch.nn.Conv2d)

    se = SqueezeExcite(latent_channels=64, min_channels=3)
    se.apply(
        event=Event.INIT,
        state=state_with_model,
        logger=logger,
    )

    return original_conv_count, original_size


def test_squeeze_excite_layer_replacement(state_with_model: State, simple_conv_model_input: Tensors,
                                          noop_dummy_logger: Logger):
    original_conv_count, _ = _do_squeeze_excite(
        state_with_model,
        simple_conv_model_input,
        noop_dummy_logger,
    )

    # verify layer replacement
    se_count = surgery.count_module_instances(state_with_model.model, SqueezeExciteConv2d)
    assert original_conv_count == se_count


def test_squeeze_excite_forward_shape(state_with_model: State, simple_conv_model_input: Tensors,
                                      noop_dummy_logger: Logger):
    _, original_size = _do_squeeze_excite(
        state_with_model,
        simple_conv_model_input,
        noop_dummy_logger,
    )

    # verify forward prop still works
    batch = (simple_conv_model_input, torch.Tensor())
    out = state_with_model.model.forward(batch)
    post_se_size = out.size()
    assert original_size == post_se_size


def test_squeeze_excite_algorithm_logging(state_with_model: State, logger_mock: Logger):
    se = SqueezeExcite(latent_channels=64, min_channels=3)
    se.apply(Event.INIT, state_with_model, logger=logger_mock)
    conv_count = surgery.count_module_instances(state_with_model.model, torch.nn.Conv2d)

    logger_mock.metric_fit.assert_called_once_with({
        'squeeze_excite/num_squeeze_excite_layers': conv_count,
    })


@pytest.mark.run_long
@pytest.mark.timeout(90)
def test_squeeze_excite_trains(mosaic_trainer_hparams: TrainerHparams):
    mosaic_trainer_hparams.algorithms = [SqueezeExciteHparams(latent_channels=32, min_channels=32)]
    train_model(mosaic_trainer_hparams, run_loss_check=True)



# Copyright 2021 MosaicML. All Rights Reserved.

from typing import Mapping, Union

import torch
from torch import Tensor
from torchmetrics import Metric

from composer.models.loss import soft_cross_entropy


class LanguageCrossEntropyLoss(Metric):
    """
    Implements a CrossEntropyLoss metric that can be run during the validation step, and is compatible with the HF API.

    Args:
        dist_sync_on_step (bool): Synchronize metric state across processes at each forward() before returning
                                  the value at the step.

    State:
        sum_loss (float): the sum of the per-example loss in the batch.
        total_batches (float): the number of batches to average across.
    """

    def __init__(self, dist_sync_on_step=False):
        super().__init__(dist_sync_on_step=dist_sync_on_step)

        self.add_state("sum_loss", default=torch.tensor(0.), dist_reduce_fx="sum")
        self.add_state("total_batches", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, output: Union[Mapping, Tensor], target: Tensor) -> None:
        """
        Updates the internal state of the Metrics object with results from a new batch.

        Args:
            output (Mapping): the output from the model, which must contain either the Tensor or a Mapping type that
                              contains the loss or model logits.
            target (Tensor): a Tensor of ground-truth values to compare against.

        Returns:
            None
        """

        # if logit modification algorithms aren't on, we take the loss directly from the model output
        if isinstance(output, Mapping) and 'loss' in output:
            loss = output['loss']
        else:
            if isinstance(output, Mapping):
                logits = output['logits']
            # recompute the loss on our own
            elif isinstance(output, Tensor):
                logits = output
            else:
                raise Exception(f"Type {type(output)} for the output is unsupported.")

            loss = soft_cross_entropy(logits, target)

        # accmulate loss over all batches
        self.sum_loss += loss

        self.total_batches += 1  #type: ignore (third-party)

    def compute(self) -> Tensor:
        """
        After all update operations, it computes the final metric from the internal state.

        Args;
            None

        Returns:
            loss (Tensor): the loss averaged across all batches.
        """
        # Return average loss over entire dataset
        return self.sum_loss / self.total_batches  #type: ignore (third-party)


class Perplexity(LanguageCrossEntropyLoss):
    """
    Subclasses the LanguageCrossEntropyLoss to provide a perplexity measurement.

    Notes:
        If an algorithm modifies the loss function and it is no longer directly provided in the output, then this
        could be expensive because it'll compute the loss twice.
    """

    def compute(self) -> Tensor:
        """
        Returns torch.exp() of the LanguageCrossEntropyLoss.
        """

        avg_loss = super().compute()
        return torch.exp(avg_loss)



# Copyright 2021 MosaicML. All Rights Reserved.

import os
from unittest.mock import patch

import pytest
import torch
from torch.optim import Adam

from composer.callbacks.lr_monitor import LRMonitor
from composer.core.logging.logger import Logger
from composer.core.precision import Precision
from composer.datasets.hparams import DataloaderSpec
from composer.loggers.tqdm_logger import TQDMLoggerBackend
from composer.models.base import BaseMosaicModel
from composer.optim.optimizer_hparams import AdamHparams
from composer.optim.scheduler import ComposedScheduler, ExponentialLRHparams
from composer.trainer import Trainer, TrainerHparams
from composer.trainer.ddp import FileStoreHparams
from composer.trainer.devices.device_hparams import CPUDeviceHparams, DeviceHparams, GPUDeviceHparams
from tests.utils.trainer_fit import get_total_loss, train_model


def test_trainer_init_all_defaults(dummy_dataloader_spec: DataloaderSpec, dummy_model: BaseMosaicModel):
    trainer = Trainer(model=dummy_model,
                      train_dataloader_spec=dummy_dataloader_spec,
                      eval_dataloader_spec=dummy_dataloader_spec,
                      max_epochs=10,
                      train_batch_size=32,
                      eval_batch_size=32)

    assert isinstance(trainer, Trainer)


def test_trainer_init_additional_args(dummy_dataloader_spec: DataloaderSpec, dummy_model: BaseMosaicModel):
    trainer = Trainer(
        model=dummy_model,
        train_dataloader_spec=dummy_dataloader_spec,
        eval_dataloader_spec=dummy_dataloader_spec,
        max_epochs=10,
        train_batch_size=32,
        eval_batch_size=32,
        optimizer_hparams=AdamHparams(),
        schedulers_hparams=[ExponentialLRHparams(gamma=0.1)],
        log_destinations=[TQDMLoggerBackend()],
        callbacks=(LRMonitor(),),
    )

    assert isinstance(trainer, Trainer)
    assert isinstance(trainer.state.optimizers, Adam)

    assert isinstance(trainer.state.schedulers, ComposedScheduler)

    assert len(trainer.logger.backends) == 1
    assert isinstance(trainer.logger.backends[0], TQDMLoggerBackend)
    assert isinstance(trainer.logger, Logger)

    # log destination and lr monitor, logger destination callback must be first
    assert len(trainer.engine.callbacks) == 2
    assert isinstance(trainer.engine.callbacks[0], TQDMLoggerBackend)
    assert isinstance(trainer.engine.callbacks[1], LRMonitor)


def test_trainer_create_from_hparams(mosaic_trainer_hparams: TrainerHparams):
    trainer = Trainer.create_from_hparams(hparams=mosaic_trainer_hparams)
    assert isinstance(trainer, Trainer)


@pytest.mark.parametrize('invalid_hparams', [])
def test_trainer_validation(mosaic_trainer_hparams: TrainerHparams, invalid_hparams):
    with patch.multiple(mosaic_trainer_hparams, **invalid_hparams), pytest.raises(ValueError):
        mosaic_trainer_hparams.validate()


@pytest.mark.run_long
@pytest.mark.timeout(90)
def test_trainer_determinism(mosaic_trainer_hparams: TrainerHparams, ddp_tmpdir: str):
    mosaic_trainer_hparams.seed = 10
    mosaic_trainer_hparams.deterministic_mode = True
    mosaic_trainer_hparams.max_epochs = 2

    mosaic_trainer_hparams.ddp.store = FileStoreHparams(os.path.join(ddp_tmpdir, "first_store"))
    first_trainer = Trainer.create_from_hparams(mosaic_trainer_hparams)
    first_trainer.fit()
    first_model = first_trainer.state.model.module
    assert isinstance(first_model, BaseMosaicModel)
    assert first_trainer.state.train_dataloader is not None
    first_loss = get_total_loss(first_model, first_trainer.state.train_dataloader)

    # Second trainer must be created after fitting the first so that the
    # seeds get fully reset for the second training run
    mosaic_trainer_hparams.ddp.store = FileStoreHparams(os.path.join(ddp_tmpdir, "second_store"))
    second_trainer = Trainer.create_from_hparams(mosaic_trainer_hparams)
    second_trainer.fit()
    second_model = second_trainer.state.model.module
    assert isinstance(second_model, BaseMosaicModel)
    assert second_trainer.state.train_dataloader is not None
    second_loss = get_total_loss(second_model, second_trainer.state.train_dataloader)

    torch.testing.assert_allclose(second_loss, first_loss)


@pytest.mark.run_long
@pytest.mark.timeout(90)
@pytest.mark.parametrize("device_hparams", [
    pytest.param(CPUDeviceHparams(n_cpus=1), id="1cpu"),
    pytest.param(CPUDeviceHparams(n_cpus=2), id='2cpu'),
    pytest.param(GPUDeviceHparams(n_gpus=1), marks=pytest.mark.n_gpus(1), id="1gpu"),
    pytest.param(GPUDeviceHparams(n_gpus=2), marks=pytest.mark.n_gpus(2), id="2gpu"),
])
@pytest.mark.parametrize("grad_accum", [
    pytest.param(1, id="ga1"),
    pytest.param(2, id="ga2"),
])
@pytest.mark.parametrize("precision", [
    pytest.param(Precision.FP32, id="fp32"),
    pytest.param(Precision.AMP, id="amp"),
])
def test_trainer_fit(mosaic_trainer_hparams: TrainerHparams, device_hparams: DeviceHparams, grad_accum: int,
                     precision: Precision):
    mosaic_trainer_hparams.device = device_hparams
    mosaic_trainer_hparams.grad_accum = grad_accum
    mosaic_trainer_hparams.precision = precision

    # Not supported
    if precision == Precision.AMP and isinstance(device_hparams, CPUDeviceHparams):
        return

    train_model(mosaic_trainer_hparams, max_epochs=2, run_loss_check=True)



# Copyright 2021 MosaicML. All Rights Reserved.

from __future__ import annotations

import warnings
from abc import ABC
from typing import TYPE_CHECKING, List, Optional, Tuple, final

from composer.core.callback import Callback, RankZeroCallback
from composer.core.logging.logger import Logger
from composer.utils.ddp import is_rank_set, is_rank_zero

if TYPE_CHECKING:
    from composer.core.logging.logger import LogLevel, TLogData
    from composer.core.state import State


class DeferredLogMetricWarning(UserWarning):
    pass


class BaseLoggerBackend(Callback, ABC):
    """BaseLoggerBackend defines the interface for logging destinations."""

    def will_log(self, state: State, log_level: LogLevel) -> bool:
        """
        will_log() returns whether the logging destination will log a metric
        or state change, given current state `state` and level `log_level`
        """
        return True

    def log_metric(self, epoch: int, step: int, log_level: LogLevel, data: TLogData):
        """
        log_metric() is called whenever there is a metric to log, for which `self.will_log(state, log_level)`
        returned True.
        """
        pass


class RankZeroLoggerBackend(BaseLoggerBackend, RankZeroCallback, ABC):

    def __init__(self) -> None:
        super().__init__()
        # self.deferred_log_metric_calls is set to None once the logger is initialized
        self.deferred_log_metric_calls: Optional[List[Tuple[int, int, LogLevel, TLogData]]] = []

    def _will_log(self, state: State, log_level: LogLevel) -> bool:
        """
        _will_log() returns whether the logging destination will log a metric
        or state change, given current state `state` and level `log_level`
        """
        return True

    @final
    def will_log(self, state: State, log_level: LogLevel) -> bool:
        if state.is_rank_set and not state.is_rank_zero:
            return False
        return self._will_log(state, log_level)

    def _log_metric(self, epoch: int, step: int, log_level: LogLevel, data: TLogData) -> None:
        """
        _log_metric() is called whenever there is a metric to log, for which `self.will_log(state, log_level)`
        returned True.
        """
        pass

    @final
    def log_metric(self, epoch: int, step: int, log_level: LogLevel, data: TLogData) -> None:
        if is_rank_set() and not is_rank_zero():
            # definitely not interested in the log message
            # lazily clear all deferred calls that will not be logged
            # so the memory can be freed
            self.deferred_log_metric_calls = None
            return
        if self.deferred_log_metric_calls is not None:
            warnings.warn(
                f"{self.__class__.__name__}.log_metric() was invoked before training_start()."
                "This log call will be queued and processed after training_start().",
                category=DeferredLogMetricWarning)
            self.deferred_log_metric_calls.append((epoch, step, log_level, data))
            return
        return self._log_metric(epoch, step, log_level, data)

    def _training_start(self, state: State, logger: Logger) -> None:
        pass

    @final
    def training_start(self, state: State, logger: Logger) -> None:
        self._training_start(state, logger)  # initialize the logger
        if self.deferred_log_metric_calls is None:
            raise RuntimeError("deferred_log_metric_calls should not be None")
        for epoch, step, log_level, data in self.deferred_log_metric_calls:
            self._log_metric(epoch, step, log_level, data)
        self.deferred_log_metric_calls = None



# Copyright 2021 MosaicML. All Rights Reserved.

import torch

from composer.models.resnets import Bottleneck


def _sample_drop(x: torch.Tensor, sample_drop_rate: float, is_training: bool):
    """Randomly drops samples from the input batch according to the `sample_drop_rate`.
    This is implemented by setting the samples to be dropped to zeros.
    """

    keep_probability = (1 - sample_drop_rate)
    if not is_training:
        return x * keep_probability
    rand_dim = [x.shape[0]] + [1] * len(x.shape[1:])
    sample_mask = keep_probability + torch.rand(rand_dim, dtype=x.dtype, device=x.device)
    sample_mask.floor_()  # binarize
    x *= sample_mask
    return x


class SampleStochasticBottleneck(Bottleneck):
    """Sample-wise stochastic ResNet Bottleneck block.

    This block has a probability of dropping samples before the identity
    connection, then adds back the untransformed samples using the identity
    connection.

    Args:
        sample_drop_rate (float): The probability of dropping a sample within
            this block. Must be between 0.0 and 1.0.
        **kwargs: Used for the original Bottleneck initialization arguments.
    """

    def __init__(self, drop_rate: float, **kwargs):
        super(SampleStochasticBottleneck, self).__init__(**kwargs)

        self.drop_rate = drop_rate

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        if self.training and self.drop_rate:
            out = _sample_drop(out, self.drop_rate, self.training)
        out += identity

        return self.relu(out)

    @staticmethod
    def from_target_layer(module: Bottleneck, module_index: int, module_count: int, drop_rate: float,
                          drop_distribution: str):
        """Helper function to convert a ResNet Bottleneck block into a sample-wise stochastic block."""

        if drop_distribution == 'linear':
            drop_rate = ((module_index + 1) / module_count) * drop_rate

        return SampleStochasticBottleneck(drop_rate=drop_rate,
                                          inplanes=module.conv1.in_channels,
                                          planes=module.conv3.out_channels // module.expansion,
                                          stride=module.stride,
                                          downsample=module.downsample,
                                          groups=module.conv2.groups,
                                          dilation=module.conv2.dilation)



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.utils.string_enum import StringEnum


class Precision(StringEnum):
    """Enum class for the numerical precision to be used by the model."""

    AMP = "amp"
    FP32 = "fp32"



# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.augmix.augmix import AugmentAndMixTransform as AugmentAndMixTransform
from composer.algorithms.augmix.augmix import AugMix as AugMix
from composer.algorithms.augmix.augmix import AugMixHparams as AugMixHparams
from composer.algorithms.augmix.augmix import augment_and_mix as augment_and_mix

_name = 'AugMix'
_class_name = 'AugMix'
_functional = 'augment_and_mix'
_tldr = ''
_attribution = '(Hendrycks et al, 2020)'
_link = 'http://arxiv.org/abs/1912.02781'
_method_card = ''
